import copy

import lenstronomy.Util.util as util
from lenstronomy.Util import class_creator
import matplotlib.pyplot as plt
import numpy as np
from mpl_toolkits.axes_grid1 import make_axes_locatable
from lenstronomy.LensModel.lens_model_extensions import LensModelExtensions
from lenstronomy.Data.coord_transforms import Coordinates
from lenstronomy.Plots import plot_util

__all__ = ["TracerPlot"]


class TracerPlot(object):
    """Class to plot a single band given the modeling results."""

    def __init__(
        self,
        kwargs_data_joint,
        kwargs_model,
        kwargs_params,
        kwargs_likelihood,
        arrow_size=0.02,
        cmap_string="gist_heat",
        fast_caustic=True,
    ):
        """

        :param kwargs_model: model keyword argument list for the full multi-band modeling
        :param kwargs_params: keyword argument of keyword argument lists of the different model components selected for
         the imaging band, NOT including linear amplitudes (not required as being overwritten by the param list)
        :param arrow_size: size of the scale and orientation arrow
        :param cmap_string: string of color map (or cmap matplotlib object)
        :param fast_caustic: boolean; if True, uses fast (but less accurate) caustic calculation method
        """

        multi_band_list = kwargs_data_joint.get("multi_band_list", [])
        multi_band_type = kwargs_data_joint.get("multi_band_type", "single-band")
        bands_compute = kwargs_data_joint.get("bands_compute", None)
        image_likelihood_mask_list = kwargs_likelihood.get(
            "image_likelihood_mask_list", None
        )
        self._tracer_light_model_band = kwargs_model.get("tracer_source_band", 0)
        self.image_model = class_creator.create_im_sim(
            multi_band_list,
            multi_band_type,
            kwargs_model,
            bands_compute=bands_compute,
            image_likelihood_mask_list=None,
            band_index=0,
            kwargs_pixelbased=None,
            linear_solver=True,
        )

        tracer_likelihood_mask = kwargs_likelihood.get("tracer_likelihood_mask", None)
        tracer_data = kwargs_data_joint.get("tracer_data", None)
        self.tracerModel = class_creator.create_tracer_model(
            tracer_data, kwargs_model, tracer_likelihood_mask=tracer_likelihood_mask
        )
        tracer_likelihood_mask = self.tracerModel.likelihood_mask
        kwargs_params_copy = copy.deepcopy(kwargs_params)
        kwargs_params_copy.pop("kwargs_tracer_source", None)
        logL, param = self.image_model.likelihood_data_given_model(**kwargs_params_copy)

        (
            self._kwargs_lens,
            self._kwargs_source,
            self._kwargs_lens_light,
            self._kwargs_ps,
        ) = self.image_model.update_linear_kwargs(
            param,
            model_band=self._tracer_light_model_band,
            kwargs_lens=kwargs_params.get("kwargs_lens", None),
            kwargs_source=kwargs_params.get("kwargs_source", None),
            kwargs_lens_light=kwargs_params.get("kwargs_lens_light", None),
            kwargs_ps=kwargs_params.get("kwargs_ps", None),
        )
        self._kwargs_tracer_source = kwargs_params["kwargs_tracer_source"]
        self._kwargs_special = kwargs_params.get("kwargs_special", None)
        self._kwargs_extinction = kwargs_params.get("kwargs_extinction", None)
        self.tracerModel = class_creator.create_tracer_model(
            tracer_data, kwargs_model, tracer_likelihood_mask=tracer_likelihood_mask
        )
        self._coords = self.tracerModel.Data
        self._data = self._coords.data
        self._model = self.tracerModel.tracer_model(
            kwargs_tracer_source=self._kwargs_tracer_source,
            kwargs_lens=self._kwargs_lens,
            kwargs_source=self._kwargs_source,
            kwargs_special=self._kwargs_special,
            kwargs_extinction=self._kwargs_extinction,
        )
        C_D = self._coords.C_D_model(self._model)
        self._norm_residuals = (
            (self._data - self._model) / np.sqrt(C_D) * tracer_likelihood_mask
        )
        self.LensModel = self.tracerModel.LensModel
        self._lensModelExt = LensModelExtensions(self.LensModel)
        self.PointSource = self.tracerModel.PointSource
        log_model = np.log10(self._model)
        log_model[np.isnan(log_model)] = -5
        self._v_min_default = max(np.min(log_model), -5)
        self._v_max_default = min(np.max(log_model), 10)

        self._data = self._coords.data
        self._deltaPix = self._coords.pixel_width
        self._frame_size = np.max(self._coords.width)
        x_grid, y_grid = self._coords.pixel_coordinates
        self._x_grid = util.image2array(x_grid)
        self._y_grid = util.image2array(y_grid)
        self._x_center, self._y_center = self._coords.center

        self._cmap = plot_util.cmap_conf(cmap_string)
        self._arrow_size = arrow_size
        self._fast_caustic = fast_caustic

    def _critical_curves(self):
        if not hasattr(self, "_ra_crit_list") or not hasattr(self, "_dec_crit_list"):
            if self._fast_caustic:
                (
                    self._ra_crit_list,
                    self._dec_crit_list,
                    self._ra_caustic_list,
                    self._dec_caustic_list,
                ) = self._lensModelExt.critical_curve_caustics(
                    self._kwargs_lens,
                    compute_window=self._frame_size,
                    grid_scale=self._deltaPix,
                    center_x=self._x_center,
                    center_y=self._y_center,
                )
                self._caustic_points_only = False
            else:
                # only supports individual points due to output of critical_curve_tiling definition
                self._caustic_points_only = True
                (
                    self._ra_crit_list,
                    self._dec_crit_list,
                ) = self._lensModelExt.critical_curve_tiling(
                    self._kwargs_lens,
                    compute_window=self._frame_size,
                    start_scale=self._deltaPix / 5.0,
                    max_order=10,
                    center_x=self._x_center,
                    center_y=self._y_center,
                )
                (
                    self._ra_caustic_list,
                    self._dec_caustic_list,
                ) = self.LensModel.ray_shooting(
                    self._ra_crit_list, self._dec_crit_list, self._kwargs_lens
                )
        return self._ra_crit_list, self._dec_crit_list

    def _caustics(self):
        if not hasattr(self, "_ra_caustic_list") or not hasattr(
            self, "_dec_caustic_list"
        ):
            _, _ = self._critical_curves()
        return self._ra_caustic_list, self._dec_caustic_list

    def data_plot(
        self,
        ax,
        v_min=None,
        v_max=None,
        text="Observed",
        font_size=15,
        colorbar_label=r"log$_{10}$ flux",
        **kwargs
    ):
        """

        :param ax:
        :return:
        """
        if v_min is None:
            v_min = self._v_min_default
        if v_max is None:
            v_max = self._v_max_default
        im = ax.matshow(
            np.log10(self._data),
            origin="lower",
            extent=[0, self._frame_size, 0, self._frame_size],
            cmap=self._cmap,
            vmin=v_min,
            vmax=v_max,
        )  # , vmin=0, vmax=2

        ax.get_xaxis().set_visible(False)
        ax.get_yaxis().set_visible(False)
        ax.autoscale(False)

        plot_util.scale_bar(
            ax, self._frame_size, dist=1, text='1"', font_size=font_size
        )
        plot_util.text_description(
            ax,
            self._frame_size,
            text=text,
            color="w",
            backgroundcolor="k",
            font_size=font_size,
        )

        if "no_arrow" not in kwargs or not kwargs["no_arrow"]:
            plot_util.coordinate_arrows(
                ax,
                self._frame_size,
                self._coords,
                color="w",
                arrow_size=self._arrow_size,
                font_size=font_size,
            )

        divider = make_axes_locatable(ax)
        cax = divider.append_axes("right", size="5%", pad=0.05)
        cb = plt.colorbar(im, cax=cax, orientation="vertical")
        cb.set_label(colorbar_label, fontsize=font_size)
        return ax

    def model_plot(
        self,
        ax,
        v_min=None,
        v_max=None,
        image_names=False,
        colorbar_label=r"log$_{10}$ flux",
        font_size=15,
        text="Reconstructed",
        **kwargs
    ):
        """

        :param ax: matplotib axis instance
        :param v_min:
        :param v_max:
        :return:
        """
        if v_min is None:
            v_min = self._v_min_default
        if v_max is None:
            v_max = self._v_max_default
        im = ax.matshow(
            np.log10(self._model),
            origin="lower",
            vmin=v_min,
            vmax=v_max,
            extent=[0, self._frame_size, 0, self._frame_size],
            cmap=self._cmap,
        )
        ax.get_xaxis().set_visible(False)
        ax.get_yaxis().set_visible(False)
        ax.autoscale(False)
        plot_util.scale_bar(
            ax, self._frame_size, dist=1, text='1"', font_size=font_size
        )
        plot_util.text_description(
            ax,
            self._frame_size,
            text=text,
            color="w",
            backgroundcolor="k",
            font_size=font_size,
        )
        if "no_arrow" not in kwargs or not kwargs["no_arrow"]:
            plot_util.coordinate_arrows(
                ax,
                self._frame_size,
                self._coords,
                color="w",
                arrow_size=self._arrow_size,
                font_size=font_size,
            )
        divider = make_axes_locatable(ax)
        cax = divider.append_axes("right", size="5%", pad=0.05)
        cb = plt.colorbar(im, cax=cax)
        cb.set_label(colorbar_label, fontsize=font_size)

        # plot_line_set(ax, self._coords, self._ra_caustic_list, self._dec_caustic_list, color='b')
        # plot_line_set(ax, self._coords, self._ra_crit_list, self._dec_crit_list, color='r')
        if image_names is True:
            ra_image, dec_image = self.PointSource.image_position(
                self._kwargs_ps,
                self._kwargs_lens,
                original_position=kwargs.get("original_position", True),
            )
            plot_util.image_position_plot(
                ax,
                self._coords,
                ra_image,
                dec_image,
                image_name_list=kwargs.get("image_name_list", None),
            )
        # source_position_plot(ax, self._coords, self._kwargs_source)

    def convergence_plot(
        self,
        ax,
        text="Convergence",
        v_min=None,
        v_max=None,
        font_size=15,
        colorbar_label=r"$\log_{10}\ \kappa$",
        **kwargs
    ):
        """

        :param ax: matplotib axis instance
        :return: convergence plot in ax instance
        """
        if not "cmap" in kwargs:
            kwargs["cmap"] = self._cmap

        kappa_result = util.array2image(
            self.LensModel.kappa(self._x_grid, self._y_grid, self._kwargs_lens)
        )
        im = ax.matshow(
            np.log10(kappa_result),
            origin="lower",
            extent=[0, self._frame_size, 0, self._frame_size],
            cmap=kwargs["cmap"],
            vmin=v_min,
            vmax=v_max,
        )
        ax.get_xaxis().set_visible(False)
        ax.get_yaxis().set_visible(False)
        ax.autoscale(False)
        plot_util.scale_bar(
            ax, self._frame_size, dist=1, text='1"', color="w", font_size=font_size
        )
        if "no_arrow" not in kwargs or not kwargs["no_arrow"]:
            plot_util.coordinate_arrows(
                ax,
                self._frame_size,
                self._coords,
                color="w",
                arrow_size=self._arrow_size,
                font_size=font_size,
            )
            plot_util.text_description(
                ax,
                self._frame_size,
                text=text,
                color="w",
                backgroundcolor="k",
                flipped=False,
                font_size=font_size,
            )
        divider = make_axes_locatable(ax)
        cax = divider.append_axes("right", size="5%", pad=0.05)
        cb = plt.colorbar(im, cax=cax)
        cb.set_label(colorbar_label, fontsize=font_size)
        return ax

    def normalized_residual_plot(
        self,
        ax,
        v_min=-6,
        v_max=6,
        font_size=15,
        text="Normalized Residuals",
        colorbar_label=r"(f${}_{\rm model}$ - f${}_{\rm data}$)/$\sigma$",
        no_arrow=False,
        color_bar=True,
        **kwargs
    ):
        """

        :param ax:
        :param v_min:
        :param v_max:
        :param kwargs: kwargs to send to matplotlib.pyplot.matshow()
        :param color_bar: Option to display the color bar
        :return:
        """
        if not "cmap" in kwargs:
            kwargs["cmap"] = "bwr"
        im = ax.matshow(
            self._norm_residuals,
            vmin=v_min,
            vmax=v_max,
            extent=[0, self._frame_size, 0, self._frame_size],
            origin="lower",
            **kwargs
        )
        ax.get_xaxis().set_visible(False)
        ax.get_yaxis().set_visible(False)
        ax.autoscale(False)
        plot_util.scale_bar(
            ax, self._frame_size, dist=1, text='1"', color="k", font_size=font_size
        )
        plot_util.text_description(
            ax,
            self._frame_size,
            text=text,
            color="k",
            backgroundcolor="w",
            font_size=font_size,
        )
        if not no_arrow:
            plot_util.coordinate_arrows(
                ax,
                self._frame_size,
                self._coords,
                color="w",
                arrow_size=self._arrow_size,
                font_size=font_size,
            )
        if color_bar:
            divider = make_axes_locatable(ax)
            cax = divider.append_axes("right", size="5%", pad=0.05)
            cb = plt.colorbar(im, cax=cax)
            cb.set_label(colorbar_label, fontsize=font_size)
        return ax

    def absolute_residual_plot(
        self,
        ax,
        v_min=-1,
        v_max=1,
        font_size=15,
        text="Residuals",
        colorbar_label=r"(f$_{model}$-f$_{data}$)",
    ):
        """

        :param ax:
        :return:
        """
        im = ax.matshow(
            self._model - self._data,
            vmin=v_min,
            vmax=v_max,
            extent=[0, self._frame_size, 0, self._frame_size],
            cmap="bwr",
            origin="lower",
        )
        ax.get_xaxis().set_visible(False)
        ax.get_yaxis().set_visible(False)
        ax.autoscale(False)
        plot_util.scale_bar(
            ax, self._frame_size, dist=1, text='1"', color="k", font_size=font_size
        )
        plot_util.text_description(
            ax,
            self._frame_size,
            text=text,
            color="k",
            backgroundcolor="w",
            font_size=font_size,
        )
        plot_util.coordinate_arrows(
            ax,
            self._frame_size,
            self._coords,
            font_size=font_size,
            color="k",
            arrow_size=self._arrow_size,
        )
        divider = make_axes_locatable(ax)
        cax = divider.append_axes("right", size="5%", pad=0.05)
        cb = plt.colorbar(im, cax=cax)
        cb.set_label(colorbar_label, fontsize=font_size)
        return ax

    def source(self, numPix, deltaPix, center=None, image_orientation=True):
        """

        :param numPix: number of pixels per axes
        :param deltaPix: pixel size
        :param image_orientation: bool, if True, uses frame in orientation of the image, otherwise in RA-DEC coordinates
        :return: 2d surface brightness grid of the reconstructed source and Coordinates() instance of source grid
        """
        if image_orientation is True:
            Mpix2coord = self._coords.transform_pix2angle * deltaPix / self._deltaPix
            x_grid_source, y_grid_source = util.make_grid_transformed(
                numPix, Mpix2Angle=Mpix2coord
            )
            ra_at_xy_0, dec_at_xy_0 = x_grid_source[0], y_grid_source[0]
        else:
            (
                x_grid_source,
                y_grid_source,
                ra_at_xy_0,
                dec_at_xy_0,
                x_at_radec_0,
                y_at_radec_0,
                Mpix2coord,
                Mcoord2pix,
            ) = util.make_grid_with_coordtransform(numPix, deltaPix)

        center_x = 0
        center_y = 0
        if center is not None:
            center_x, center_y = center[0], center[1]
        elif len(self._kwargs_tracer_source) > 0:
            center_x = self._kwargs_tracer_source[0]["center_x"]
            center_y = self._kwargs_tracer_source[0]["center_y"]
        x_grid_source += center_x
        y_grid_source += center_y

        coords_source = Coordinates(
            transform_pix2angle=Mpix2coord,
            ra_at_xy_0=ra_at_xy_0 + center_x,
            dec_at_xy_0=dec_at_xy_0 + center_y,
        )

        source = self.tracerModel.tracer_source_class.surface_brightness(
            x_grid_source, y_grid_source, self._kwargs_tracer_source
        )
        source = util.array2image(source)
        return source, coords_source

    def source_plot(
        self,
        ax,
        numPix,
        deltaPix_source,
        center=None,
        v_min=None,
        v_max=None,
        with_caustics=False,
        caustic_color="yellow",
        font_size=15,
        plot_scale="log",
        scale_size=0.1,
        text="Reconstructed source",
        colorbar_label=r"tracer",
        point_source_position=True,
        **kwargs
    ):
        """

        :param ax:
        :param numPix:
        :param deltaPix_source:
        :param center: [center_x, center_y], if specified, uses this as the center
        :param v_min:
        :param v_max:
        :param caustic_color:
        :param font_size:
        :param plot_scale: string, log or linear, scale of surface brightness plot
        :param kwargs:
        :return:
        """
        if v_min is None:
            v_min = self._v_min_default
        if v_max is None:
            v_max = self._v_max_default
        d_s = numPix * deltaPix_source
        source, coords_source = self.source(numPix, deltaPix_source, center=center)
        if plot_scale == "log":
            source[source < 10**v_min] = 10 ** (v_min)  # to remove weird shadow in plot
            source_scale = np.log10(source)
        elif plot_scale == "linear":
            source_scale = source
        else:
            raise ValueError(
                'variable plot_scale needs to be "log" or "linear", not %s.'
                % plot_scale
            )
        im = ax.matshow(
            source_scale,
            origin="lower",
            extent=[0, d_s, 0, d_s],
            cmap=self._cmap,
            vmin=v_min,
            vmax=v_max,
        )  # source
        ax.get_xaxis().set_visible(False)
        ax.get_yaxis().set_visible(False)
        ax.autoscale(False)
        divider = make_axes_locatable(ax)
        cax = divider.append_axes("right", size="5%", pad=0.05)
        cb = plt.colorbar(im, cax=cax)
        cb.set_label(colorbar_label, fontsize=font_size)

        if with_caustics is True:
            ra_caustic_list, dec_caustic_list = self._caustics()
            plot_util.plot_line_set(
                ax,
                coords_source,
                ra_caustic_list,
                dec_caustic_list,
                color=caustic_color,
                points_only=self._caustic_points_only,
            )
            plot_util.plot_line_set(
                ax,
                coords_source,
                ra_caustic_list,
                dec_caustic_list,
                color=caustic_color,
                points_only=self._caustic_points_only,
                **kwargs.get("kwargs_caustic", {})
            )
            plot_util.scale_bar(
                ax,
                d_s,
                dist=scale_size,
                text='{:.1f}"'.format(scale_size),
                color="w",
                flipped=False,
                font_size=font_size,
            )
        if "no_arrow" not in kwargs or not kwargs["no_arrow"]:
            plot_util.coordinate_arrows(
                ax,
                self._frame_size,
                self._coords,
                color="w",
                arrow_size=self._arrow_size,
                font_size=font_size,
            )
            plot_util.text_description(
                ax,
                d_s,
                text=text,
                color="w",
                backgroundcolor="k",
                flipped=False,
                font_size=font_size,
            )
        if point_source_position is True:
            ra_source, dec_source = self.PointSource.source_position(
                self._kwargs_ps, self._kwargs_lens
            )
            plot_util.source_position_plot(ax, coords_source, ra_source, dec_source)
        return ax

    def magnification_plot(
        self,
        ax,
        v_min=-10,
        v_max=10,
        image_name_list=None,
        font_size=15,
        no_arrow=False,
        text="Magnification model",
        colorbar_label=r"$\det\ (\mathsf{A}^{-1})$",
        **kwargs
    ):
        """

        :param ax: matplotib axis instance
        :param v_min: minimum range of plotting
        :param v_max: maximum range of plotting
        :param kwargs: kwargs to send to matplotlib.pyplot.matshow()
        :return:
        """
        if "cmap" not in kwargs:
            kwargs["cmap"] = self._cmap
        if "alpha" not in kwargs:
            kwargs["alpha"] = 0.5
        mag_result = util.array2image(
            self.LensModel.magnification(self._x_grid, self._y_grid, self._kwargs_lens)
        )
        im = ax.matshow(
            mag_result,
            origin="lower",
            extent=[0, self._frame_size, 0, self._frame_size],
            vmin=v_min,
            vmax=v_max,
            **kwargs
        )
        ax.get_xaxis().set_visible(False)
        ax.get_yaxis().set_visible(False)
        ax.autoscale(False)
        plot_util.scale_bar(
            ax, self._frame_size, dist=1, text='1"', color="k", font_size=font_size
        )
        if not no_arrow:
            plot_util.coordinate_arrows(
                ax,
                self._frame_size,
                self._coords,
                color="k",
                arrow_size=self._arrow_size,
                font_size=font_size,
            )
        plot_util.text_description(
            ax,
            self._frame_size,
            text=text,
            color="k",
            backgroundcolor="w",
            font_size=font_size,
        )
        divider = make_axes_locatable(ax)
        cax = divider.append_axes("right", size="5%", pad=0.05)
        cb = plt.colorbar(im, cax=cax)
        cb.set_label(colorbar_label, fontsize=font_size)
        ra_image, dec_image = self.PointSource.image_position(
            self._kwargs_ps, self._kwargs_lens
        )
        plot_util.image_position_plot(
            ax,
            self._coords,
            ra_image,
            dec_image,
            color="k",
            image_name_list=image_name_list,
        )
        return ax

    def deflection_plot(
        self,
        ax,
        v_min=None,
        v_max=None,
        axis=0,
        with_caustics=False,
        image_name_list=None,
        text="Deflection model",
        font_size=15,
        colorbar_label=r"arcsec",
    ):
        """

        :return:
        """

        alpha1, alpha2 = self.LensModel.alpha(
            self._x_grid, self._y_grid, self._kwargs_lens
        )
        alpha1 = util.array2image(alpha1)
        alpha2 = util.array2image(alpha2)
        if axis == 0:
            alpha = alpha1
        else:
            alpha = alpha2
        im = ax.matshow(
            alpha,
            origin="lower",
            extent=[0, self._frame_size, 0, self._frame_size],
            vmin=v_min,
            vmax=v_max,
            cmap=self._cmap,
            alpha=0.5,
        )
        ax.get_xaxis().set_visible(False)
        ax.get_yaxis().set_visible(False)
        ax.autoscale(False)
        plot_util.scale_bar(
            ax, self._frame_size, dist=1, text='1"', color="k", font_size=font_size
        )
        plot_util.coordinate_arrows(
            ax,
            self._frame_size,
            self._coords,
            color="k",
            arrow_size=self._arrow_size,
            font_size=font_size,
        )
        plot_util.text_description(
            ax,
            self._frame_size,
            text=text,
            color="k",
            backgroundcolor="w",
            font_size=font_size,
        )
        divider = make_axes_locatable(ax)
        cax = divider.append_axes("right", size="5%", pad=0.05)
        cb = plt.colorbar(im, cax=cax)
        cb.set_label(colorbar_label, fontsize=font_size)
        if with_caustics is True:
            ra_crit_list, dec_crit_list = self._critical_curves()
            ra_caustic_list, dec_caustic_list = self._caustics()
            plot_util.plot_line_set(
                ax,
                self._coords,
                ra_caustic_list,
                dec_caustic_list,
                color="b",
                points_only=self._caustic_points_only,
            )
            plot_util.plot_line_set(
                ax,
                self._coords,
                ra_crit_list,
                dec_crit_list,
                color="r",
                points_only=self._caustic_points_only,
            )
        ra_image, dec_image = self.PointSource.image_position(
            self._kwargs_ps, self._kwargs_lens
        )
        plot_util.image_position_plot(
            ax, self._coords, ra_image, dec_image, image_name_list=image_name_list
        )
        return ax

    def plot_main(self, with_caustics=False):
        """Print the main plots together in a joint frame.

        :return:
        """

        f, axes = plt.subplots(2, 3, figsize=(16, 8))
        self.data_plot(ax=axes[0, 0])
        self.model_plot(ax=axes[0, 1], image_names=True)
        self.normalized_residual_plot(ax=axes[0, 2], v_min=-6, v_max=6)
        self.source_plot(
            ax=axes[1, 0], deltaPix_source=0.01, numPix=100, with_caustics=with_caustics
        )
        self.convergence_plot(ax=axes[1, 1], v_max=1)
        self.magnification_plot(ax=axes[1, 2])
        f.tight_layout()
        f.subplots_adjust(
            left=None, bottom=None, right=None, top=None, wspace=0.0, hspace=0.05
        )
        return f, axes



from lenstronomy.Cosmo.micro_lensing import einstein_radius, source_size
from lenstronomy.LensModel.lens_model import LensModel
import numpy as np
import math


def splitting_centers(x, y, side_length, n_p):
    """Takes square centered at center = (x, y) with side_length = side length.

       (float) and divides it into n_p (integer) squares along each axis - so
       returns n_p * n_p subsquares from the original square. In particular,
       the function returns the coordinates of the centers of these subsquares,
       along with the final side length of the subsquare.

    :param x: x coordinate of center of square
    :type x: nparray
    :param y: y coordinate of center of square
    :type y: nparray
    :param side_length: side length of square
    :type side_length: float
    :param n_p: number of squares along each axis
    :type n_p: int
    :return: x and y coordinates of centers of subsquares
    :rtype: nparray
    :return: side length of subsquares
    :rtype: float
    """

    new_x = np.empty(n_p**2 * len(x))
    new_y = np.empty(n_p**2 * len(y))

    k = 0
    for i in range(n_p):
        for j in range(n_p):
            new_x[k :: n_p**2] = x - side_length / 2 + (j + 0.5) * side_length / n_p
            new_y[k :: n_p**2] = y - side_length / 2 + (i + 0.5) * side_length / n_p
            k += 1

    new_side_length = side_length / n_p
    return new_x, new_y, new_side_length


# defines loop_information to defines number of iterations and final scale factor
def loop_information(eta, beta_0, beta_s):

    N = 1 + math.log((beta_0 / beta_s), eta)
    number_of_iterations = math.ceil(N)
    N_star = N - math.floor(N)
    final_eta = eta**N_star

    return number_of_iterations, final_eta


def within_distance(x, y, test_point, threshold, threshold_2):
    """Check if points in center_points are within a threshold distance of the
    test_point, component-wise.

    :param x: x source coordrate of grid
    :type x: float
    :param y: y source coordrate of grid
    :type y: float
    :param test_point: Coordinates of the test point (x, y). Source position
    :type test_point: nparray
    :param threshold: delta_beta value.
    :type threshold: float
    :param threshold_2: beta_s value.
    :type threshold: float
    :return: Boolean array indicating whether each point is within both threshold and
        threshold_2 distances.
    :rtype: nparray
    """

    distances = ((x - test_point[0]) ** 2 + (y - test_point[1]) ** 2) ** (1 / 2)
    return (distances < threshold) & (distances < threshold_2)


# temporary function name, basically its the ABM algorithm with pixel division
def adaptive_boundary_mesh(
    source_position,
    L,
    beta_0,
    beta_s,
    n_p,
    eta,
    number_of_iterations,
    final_eta,
    kwargs_lens,
):
    """

    Original ABM algorithm described in Meena et al. 2022, https://arxiv.org/abs/2203.08131

    Returns list of those high resolution image-plane pixels that were mapped to
    within the radius :math:`\\beta_s` around the source position (:math:`\\beta_1`, :math:`\\beta_2`) in the source plane. This
    is done by first loading all image-plane pixels, ray shooting from the image plane
    to the source plane, and then checking if they are within the radius :math:`\\beta_s`.

     :param source_position: Coordinates of the source position (x, y). Source position
     :type source_position: tuple
     :param L: Side length of square area in image plane. Same as lenstronomy grid width
     :type L: float
     :param beta_0: Initial search radius (delta_beta)
     :type beta_0: float
     :param beta_s: Factor of 1/2 because radius
     :type beta_s: float
     :param n_p: Number of pixels
     :type n_p: int
     :param eta: Factor by which the search radius is reduced in each iteration
     :type eta: float
     :param number_of_iterations: Number of iterations
     :type number_of_iterations: int
     :param final_eta: Final scale factor
     :type final_eta: float
     :param kwargs_lens: Keyword arguments for lens model
     :type kwargs_lens: dict
     :return: side_length: updated side length of square area in image plane
     :rtype: side_length: nparray
     :return: total_number_of_rays_shot: total number of rays shot
     :rtype: total_number_of_rays_shot: int
     :return: subset_image_centers_x, subset_image_centers_y: x and y coordinates of high resolution image-plane pixels that were mapped to within the radius β_s around the source position (β1, β2) in the source plane
     :rtype: subset_image_centers_x, subset_image_centers_y: nparray
    """
    # default values

    # define the microlens
    d_l = 4000  # distance of the lens in pc
    d_s = 8000  # distance of the source in pc
    M0 = 0.01  # mass of the lens in units of M_sol (limited up to 0.25 M_sol)
    diameter_s = (
        20  # size of the diameter of the source star in units of the solar radius
    )
    theta_E = einstein_radius(M0, d_l, d_s)  # returns Einstein radius in arcseconds
    source_diameter = source_size(
        diameter_s, d_s
    )  # converts source star diameter from solar radius to arcseconds
    lens = LensModel(lens_model_list=["POINT_MASS"])  # define the lens model
    kwargs_lens = [
        {"theta_E": theta_E, "center_x": theta_E / 4, "center_y": theta_E / 6}
    ]

    # define square area and search radius
    L = (
        theta_E * 4
    )  # side length of square area in image plane - same as lenstronomy grid width, arc seconds
    beta_0 = (
        4 * L
    )  # initial search radius few times bigger than "necessary" to be safe, arc seconds
    beta_s = (
        source_diameter / 2
    )  # source size, factor of 1/2 because radius, arc seconds
    n_p = 5  # number of subsquares per side by which each (valid) pixel is divided in the next iteration
    eta = (
        0.7 * n_p
    )  # the factor by which the search radius is reduced in each iteration
    source_position = (0, 0)  # position of the source in the source plane

    # creates an loop_info array to store number of iterations and final scale factor
    loop_info = loop_information(eta, beta_0, beta_s)
    number_of_iterations = loop_info[0]
    final_eta = loop_info[1]

    # Initialize variables
    total_number_of_rays_shot = 0  # Counter for total number of rays shot
    i = 2  # Iteration counter (starts at 2 because we already have the source position)
    image_centers = np.array([[0, 0]])  # Initial center coordinates
    image_centers_x, image_centers_y = image_centers[:, 0], image_centers[:, 1]
    side_length = L  # Initial side length of square region (for source image), same as image plane side length, arc seconds
    delta_beta = beta_0  # Initial step size for source plane radius, arc seconds

    # Main loop for adaptive boundary mesh algorithm
    while i <= number_of_iterations:

        # Split image plane centers
        image_centers_x, image_centers_y, _ = splitting_centers(
            image_centers_x, image_centers_y, side_length, n_p
        )

        # Ray shoot from image to source plane using array-based approach
        source_centers_x, source_centers_y = lens.ray_shooting(
            image_centers_x, image_centers_y, kwargs=kwargs_lens
        )

        within_radius = within_distance(
            source_centers_x, source_centers_y, source_position, delta_beta, beta_s
        )
        subset_image_centers_x = image_centers_x[within_radius]
        subset_image_centers_y = image_centers_y[within_radius]

        # Update total number of rays shot
        total_number_of_rays_shot += len(subset_image_centers_x)

        # Update centers to be matched image centers with subset of centers in source plane
        combined_images = np.column_stack(
            (subset_image_centers_x, subset_image_centers_y)
        )  # not used in calculation, used for debugging

        # Update side length
        side_length /= n_p

        # Update delta_beta based on iteration number
        if i < number_of_iterations:
            delta_beta /= eta
        elif i == number_of_iterations:
            delta_beta /= final_eta

        # Increment iteration counter
        i += 1

    return (
        side_length,
        total_number_of_rays_shot,
        subset_image_centers_x,
        subset_image_centers_y,
    )



from lenstronomy.Plots.tracer_plot import TracerPlot
import matplotlib
import numpy as np

matplotlib.use("agg")
import matplotlib.pyplot as plt


class TestTracerPlot(object):
    def setup_method(self):
        # imagng data specifics
        background_rms = 0.005  # background noise per pixel
        exp_time = 500.0  # exposure time (arbitrary units, flux per pixel is in units #photons/exp_time unit)
        numPix = 60  # cutout pixel size per axis
        pixel_scale = 0.05  # pixel size in arcsec (area per pixel = pixel_scale**2)
        fwhm = 0.05  # full width at half maximum of PSF
        psf_type = "GAUSSIAN"  # 'GAUSSIAN', 'PIXEL', 'NONE'

        # tracer measurements specifics
        tracer_noise_map = np.ones(
            (numPix, numPix)
        )  # variance of metallicity measurement for each pixel

        # lensing quantities
        lens_model_list = ["SIE", "SHEAR"]
        kwargs_sie = {
            "theta_E": 0.66,
            "center_x": 0.05,
            "center_y": 0,
            "e1": 0.07,
            "e2": -0.03,
        }  # parameters of the deflector lens model
        kwargs_shear = {
            "gamma1": 0.0,
            "gamma2": -0.05,
        }  # shear values to the source plane

        kwargs_lens = [kwargs_sie, kwargs_shear]
        from lenstronomy.LensModel.lens_model import LensModel

        lens_model_class = LensModel(lens_model_list)

        # Sersic parameters in the initial simulation for the source
        kwargs_sersic = {
            "amp": 16,
            "R_sersic": 0.1,
            "n_sersic": 1,
            "e1": -0.1,
            "e2": 0.1,
            "center_x": 0.1,
            "center_y": 0,
        }
        source_model_list = ["SERSIC_ELLIPSE"]
        kwargs_source = [kwargs_sersic]

        from lenstronomy.LightModel.light_model import LightModel

        source_model_class = LightModel(source_model_list)

        # Source tracer model
        tracer_source_model_list = ["LINEAR"]
        kwargs_tracer_source = [
            {
                "amp": 100,
                "k": 1,
                "center_x": kwargs_sersic["center_x"],
                "center_y": kwargs_sersic["center_y"],
            }
        ]
        tracer_source_class = LightModel(tracer_source_model_list)

        # lens light model
        kwargs_sersic_lens = {
            "amp": 16,
            "R_sersic": 0.6,
            "n_sersic": 2,
            "e1": -0.1,
            "e2": 0.1,
            "center_x": 0.05,
            "center_y": 0,
        }

        lens_light_model_list = ["SERSIC_ELLIPSE"]
        kwargs_lens_light = [kwargs_sersic_lens]
        lens_light_model_class = LightModel(lens_light_model_list)

        # import main simulation class of lenstronomy
        from lenstronomy.Util import util
        from lenstronomy.Data.imaging_data import ImageData
        from lenstronomy.Data.psf import PSF
        import lenstronomy.Util.image_util as image_util
        from lenstronomy.ImSim.image_model import ImageModel

        # generate the coordinate grid and image properties (we only read out the relevant lines we need)
        (
            _,
            _,
            ra_at_xy_0,
            dec_at_xy_0,
            _,
            _,
            Mpix2coord,
            _,
        ) = util.make_grid_with_coordtransform(
            numPix=numPix,
            deltapix=pixel_scale,
            center_ra=0,
            center_dec=0,
            subgrid_res=1,
            inverse=False,
        )

        kwargs_image_data = {
            "background_rms": background_rms,  # rms of background noise
            "exposure_time": exp_time,  # exposure time (or a map per pixel)
            "ra_at_xy_0": ra_at_xy_0,  # RA at (0,0) pixel
            "dec_at_xy_0": dec_at_xy_0,  # DEC at (0,0) pixel
            "transform_pix2angle": Mpix2coord,
            # matrix to translate shift in pixel in shift in relative RA/DEC (2x2 matrix). Make sure it's units are arcseconds or the angular units you want to model.
            "image_data": np.zeros((numPix, numPix)),
            # 2d data vector, here initialized with zeros as place holders that get's overwritten once a simulated image with noise is created.
        }

        image_data_class = ImageData(**kwargs_image_data)
        # generate the psf variables
        kwargs_psf = {
            "psf_type": "GAUSSIAN",
            "fwhm": fwhm,
            "pixel_size": pixel_scale,
            "truncation": 3,
        }

        # if you are using a PSF estimate from e.g. a star in the FoV of your exposure, you can set
        # kwargs_psf = {'psf_type': 'PIXEL', 'pixel_size': deltaPix, 'kernel_point_source': 'odd numbered 2d grid with centered star/PSF model'}

        psf_class = PSF(**kwargs_psf)
        kwargs_numerics = {
            "supersampling_factor": 1,
            "supersampling_convolution": False,
        }

        imageModel = ImageModel(
            image_data_class,
            psf_class,
            lens_model_class=lens_model_class,
            source_model_class=source_model_class,
            lens_light_model_class=lens_light_model_class,
            kwargs_numerics=kwargs_numerics,
        )

        # generate image
        image_model = imageModel.image(
            kwargs_lens,
            kwargs_source,
            kwargs_lens_light=kwargs_lens_light,
            kwargs_ps=None,
        )

        poisson = image_util.add_poisson(image_model, exp_time=exp_time)
        bkg = image_util.add_background(image_model, sigma_bkd=background_rms)
        image_real = image_model + poisson + bkg

        image_data_class.update_data(image_real)
        kwargs_image_data["image_data"] = image_real

        # Tracer data

        kwargs_tracer_data = {
            "noise_map": tracer_noise_map,  # variance of pixels
            "ra_at_xy_0": ra_at_xy_0,  # RA at (0,0) pixel
            "dec_at_xy_0": dec_at_xy_0,  # DEC at (0,0) pixel
            "transform_pix2angle": Mpix2coord,
            # matrix to translate shift in pixel in shift in relative RA/DEC (2x2 matrix). Make sure it's units are arcseconds or the angular units you want to model.
            "image_data": np.zeros((numPix, numPix)),
            # 2d data vector, here initialized with zeros as place holders that get's overwritten once a simulated image with noise is created.
        }

        tracer_data_class = ImageData(**kwargs_tracer_data)

        from lenstronomy.ImSim.tracer_model import TracerModelSource

        tracer_model_class = TracerModelSource(
            data_class=tracer_data_class,
            tracer_source_class=tracer_source_class,
            psf_class=psf_class,
            lens_model_class=lens_model_class,
            source_model_class=source_model_class,
            lens_light_model_class=lens_light_model_class,
        )

        tracer_model = tracer_model_class.tracer_model(
            kwargs_tracer_source, kwargs_lens, kwargs_source
        )

        # add noise
        tracer_noise = image_util.add_background(
            image_model, sigma_bkd=tracer_noise_map
        )
        tracer_real = tracer_model + tracer_noise

        tracer_data_class.update_data(tracer_real)
        kwargs_tracer_data["image_data"] = tracer_real

        kwargs_likelihood = {"source_marg": False}
        kwargs_model = {
            "lens_model_list": lens_model_list,
            "source_light_model_list": source_model_list,
            "lens_light_model_list": lens_light_model_list,
            "tracer_source_model_list": tracer_source_model_list,
            "tracer_source_band": 0,
        }  # what imaging band's surface brightness solution is used for tracer models

        multi_band_list = [[kwargs_image_data, kwargs_psf, kwargs_numerics]]
        # if you have multiple  bands to be modeled simultaneously, you can append them to the mutli_band_list
        kwargs_data_joint = {
            "multi_band_list": multi_band_list,
            "multi_band_type": "single-band",
            # 'multi-linear': every imaging band has independent solutions of the surface brightness, 'joint-linear': there is one joint solution of the linear coefficients demanded across the bands.
            "tracer_data": [kwargs_tracer_data, kwargs_psf, kwargs_numerics],
        }
        self.kwargs_data_joint = kwargs_data_joint
        self.kwargs_model = kwargs_model
        self.kwargs_likelihood = kwargs_likelihood
        self.kwargs_params = {
            "kwargs_lens": kwargs_lens,
            "kwargs_source": kwargs_source,
            "kwargs_lens_light": kwargs_lens_light,
            "kwargs_tracer_source": kwargs_tracer_source,
        }

    def test_tracer_plot(self):
        tracer_plot = TracerPlot(
            self.kwargs_data_joint,
            self.kwargs_model,
            self.kwargs_params,
            self.kwargs_likelihood,
            arrow_size=0.02,
            cmap_string="gist_heat",
            fast_caustic=True,
        )

        f, axes = plt.subplots(2, 3, figsize=(16, 8))

        tracer_plot.data_plot(ax=axes[0, 0])
        tracer_plot.model_plot(ax=axes[0, 1])
        tracer_plot.normalized_residual_plot(ax=axes[0, 2], v_min=-6, v_max=6)
        tracer_plot.source_plot(
            ax=axes[1, 0],
            deltaPix_source=0.01,
            numPix=100,
            plot_scale="log",
            v_min=None,
            v_max=None,
        )
        tracer_plot.convergence_plot(ax=axes[1, 1], v_max=1)
        tracer_plot.magnification_plot(ax=axes[1, 2])
        f.tight_layout()
        f.subplots_adjust(
            left=None, bottom=None, right=None, top=None, wspace=0.0, hspace=0.05
        )
        plt.close()



__author__ = "ajshajib"

from lenstronomy.LensModel.Profiles.base_profile import LensProfileBase
import numpy as np

__all__ = ["BlankPlane"]


class BlankPlane(LensProfileBase):
    """Class for a blank lens plane.

    This is needed to create a blank plane that has a source without any lensing effect,
    when distance ratios are sampled in multi-lens-plane and multi-source plane case.
    """

    param_names = []
    lower_limit_default = {}
    upper_limit_default = {}

    def function(self, x, y):
        """"""
        return np.zeros_like(x)

    def derivatives(self, x, y):
        """

        :param x: coordinate in image plane (angle)
        :param y: coordinate in image plane (angle)
        :param alpha_x: shift in x-direction (angle)
        :param alpha_y: shift in y-direction (angle)
        :return: deflection in x- and y-direction
        """
        return np.zeros_like(x), np.zeros_like(x)

    def hessian(self, x, y):
        """

        :param x: coordinate in image plane (angle)
        :param y: coordinate in image plane (angle)
        :param alpha_x: shift in x-direction (angle)
        :param alpha_y: shift in y-direction (angle)
        :return: hessian elements f_xx, f_xy, f_yx, f_yy
        """
        return np.zeros_like(x), np.zeros_like(x), np.zeros_like(x), np.zeros_like(x)



import numpy as np
from lenstronomy.LensModel.Profiles.base_profile import LensProfileBase
from lenstronomy.Util import param_util
from scipy.interpolate import interp1d
from scipy import integrate

__all__ = ["RadialInterpolate"]


class RadialInterpolate(LensProfileBase):
    """Radially interpolated profile with azimuthal symmetry."""

    param_names = [
        "r_bin",
        "kappa_r",
        "center_x",
        "center_y",
    ]
    lower_limit_default = {}
    upper_limit_default = {}

    def function(
        self,
        x,
        y,
        r_bin=None,
        kappa_r=None,
        center_x=0,
        center_y=0,
    ):
        """Lensing potential (only needed for specific calculations, such as time
        delays)

        :param x: x-coordinate (angular position), float or numpy array
        :param y: y-coordinate (angular position), float or numpy array
        :param r_bin: radial bins for which convergence values are provided
        :type r_bin: array
        :param kappa_r: convergence values corresponding to the r_bin radii
        :type kappa_r: array of same size as r_bin
        :param center_x: x-position of center of radial density profile
        :type center_x: float
        :param center_y: y-position of center of radial density profile
        :type center_y: float
        :return: lensing potential
        """
        r, phi = param_util.cart2polar(x, y, center_x=center_x, center_y=center_y)
        # -\int alpha(r) dr  from 0 to r
        pot = self._potential_r(r, r_bin, kappa_r)
        return pot

    def derivatives(
        self,
        x,
        y,
        r_bin=None,
        kappa_r=None,
        center_x=0,
        center_y=0,
    ):
        """Returns df/dx and df/dy of the function.

        :param x: x-coordinate (angular position), float or numpy array
        :param y: y-coordinate (angular position), float or numpy array
        :param r_bin: radial bins for which convergence values are provided
        :type r_bin: array
        :param kappa_r: convergence values corresponding to the r_bin radii
        :type kappa_r: array of same size as r_bin
        :param center_x: x-position of center of radial density profile
        :type center_x: float
        :param center_y: y-position of center of radial density profile
        :type center_y: float
        :return: f_x, f_y at interpolated positions (x, y)
        """
        x_ = x - center_x
        y_ = y - center_y
        r = np.maximum(np.sqrt(x_**2 + y_**2), 10 ** (-10))
        alpha = self.alpha(r, r_bin, kappa_r)
        return alpha * x_ / r, alpha * y_ / r

    def hessian(
        self,
        x,
        y,
        r_bin=None,
        kappa_r=None,
        center_x=0,
        center_y=0,
    ):
        """Returns Hessian matrix of function d^2f/dx^2, d^2/dxdy, d^2/dydx, d^f/dy^2.

        :param x: x-coordinate (angular position), float or numpy array
        :param y: y-coordinate (angular position), float or numpy array
        :param r_bin: radial bins for which convergence values are provided
        :type r_bin: array
        :param kappa_r: convergence values corresponding to the r_bin radii
        :type kappa_r: array of same size as r_bin
        :param center_x: x-position of center of radial density profile
        :type center_x: float
        :param center_y: y-position of center of radial density profile
        :type center_y: float
        :return: f_xx, f_xy, f_yx, f_yy at interpolated positions (x, y)
        """
        r, phi = param_util.cart2polar(x, y, center_x=center_x, center_y=center_y)
        r = np.maximum(r, 10 ** (-10))
        kappa = self._kappa_r_interp(r, r_bin, kappa_r)
        # shear in the spherical case is the average convergence enclosed minus the convergence at the radius
        # source: Kaiser 1995
        gamma = 1 * (self._mass_enclosed(r, r_bin, kappa) / (np.pi * r**2) - kappa)
        # turn in to vector for gamma and cartesian derivatives
        gamma1 = -np.cos(2.0 * phi) * gamma
        gamma2 = -np.sin(2.0 * phi) * gamma
        f_xx = kappa + gamma1
        f_yy = kappa - gamma1
        f_xy = gamma2

        return f_xx, f_xy, f_xy, f_yy

    def alpha(self, r, r_bin, kappa_r):
        """Radial deflection angle m(<r) / r / pi.

        :param r: radius from center
        :param r_bin: radial bins for which convergence values are provided
        :type r_bin: array
        :param kappa_r: convergence values corresponding to the r_bin radii
        :type kappa_r: array of same size as r_bin
        :return: radial deflection angle
        """
        return self._mass_enclosed_over_r(r, r_bin, kappa_r) / np.pi

    def _kappa_r_interp(self, r, r_bin, kappa_r):
        """Calls interpolated kappa(r)

        :param r: radius
        :param r_bin: radial bins for which convergence values are provided
        :type r_bin: numpy array
        :param kappa_r: convergence values corresponding to the r_bin radii
        :type kappa_r: array of same size as r_bin
        :return: kappa(r)
        """
        if not hasattr(self, "_interp_kappa"):
            self._interp_kappa = interp1d(
                r_bin,
                kappa_r,
                fill_value=(kappa_r[0], kappa_r[-1]),
                bounds_error=False,
                kind="linear",
            )
        return self._interp_kappa(r)

    def _mass_enclosed(self, r, r_bin, kappa_r):
        """Convergence enclosed a radius.

        :param r: radius
        :param r_bin: radial bins for which convergence values are provided
        :type r_bin: numpy array
        :param kappa_r: convergence values corresponding to the r_bin radii
        :type kappa_r: array of same size as r_bin
        :return: integrated convergence within radius <r
        """
        if not hasattr(self, "_interp_m_enclosed"):
            m_r = self._m_r(r_bin, kappa_r)

            self._interp_m_enclosed = interp1d(
                r_bin, m_r, fill_value=(0, m_r[-1]), bounds_error=False
            )
        return self._interp_m_enclosed(r)

    def _mass_enclosed_over_r(self, r, r_bin, kappa_r):
        """Convergence enclosed a radius divided by r.

        :param r: radius
        :param r_bin: radial bins for which convergence values are provided
        :type r_bin: numpy array
        :param kappa_r: convergence values corresponding to the r_bin radii
        :type kappa_r: array of same size as r_bin
        :return: integrated convergence within radius <r
        """
        if not hasattr(self, "_interp_m_enclosed_over_r"):
            m_r = self._m_r(r_bin, kappa_r)
            r_bin_ = np.maximum(r_bin, 10 ** (-10))
            self._interp_m_enclosed_over_r = interp1d(
                r_bin,
                m_r / r_bin_,
                fill_value=(0, m_r[-1] / r_bin_[-1]),
                bounds_error=False,
                kind="linear",
            )
        return self._interp_m_enclosed_over_r(r)

    def _m_r(self, r_bin, kappa_r):
        """

        :param r_bin: radial bins for which convergence values are provided
        :type r_bin: numpy array
        :param kappa_r: convergence values corresponding to the r_bin radii
        :type kappa_r: array of same size as r_bin
        :return: integrated convergence within radius <r
        """

        def _integrand(x):
            return x * 2 * np.pi * self._kappa_r_interp(x, r_bin, kappa_r)

        m_slice_list = []
        r_min = 0
        for r_ in r_bin:
            m_slice, _ = integrate.quad(_integrand, r_min, r_)
            m_slice_list.append(m_slice)
            r_min = r_
        m_r = np.cumsum(m_slice_list)
        return m_r

    def _potential_r(self, r, r_bin, kappa_r):
        """Convergence enclosed a radius.

        :param r: radius
        :param r_bin: radial bins for which convergence values are provided
        :type r_bin: numpy array
        :param kappa_r: convergence values corresponding to the r_bin radii
        :type kappa_r: array of same size as r_bin
        :return: integrated convergence within radius <r
        """
        if not hasattr(self, "_interp_potential"):

            def _integrand(x):
                return self.alpha(x, r_bin, kappa_r)

            pot_slice_list = []
            r_min = 0
            for r_ in r_bin:
                pot_slice, _ = integrate.quad(_integrand, r_min, r_)
                pot_slice_list.append(pot_slice)
                r_min = r_
            pot_r = np.cumsum(pot_slice_list)
            r_bin_ = np.maximum(r_bin, 10 ** (-10))
            self._interp_potential_over_r2 = interp1d(
                r_bin,
                pot_r / r_bin_**2,
                fill_value=(0, pot_r[-1]),
                bounds_error=False,
            )
        return self._interp_potential_over_r2(r) * r**2



__author__ = "sibirrer"

import numpy.testing as npt
import numpy as np
import pytest

import lenstronomy.Util.param_util as param_util
from lenstronomy.LensModel.lens_model import LensModel
from lenstronomy.LightModel.light_model import LightModel
from lenstronomy.PointSource.point_source import PointSource
from lenstronomy.ImSim.image_model import ImageModel
from lenstronomy.ImSim.image_linear_solve import ImageLinearFit
import lenstronomy.Util.simulation_util as sim_util
from lenstronomy.LensModel.Solver.lens_equation_solver import LensEquationSolver
from lenstronomy.Data.imaging_data import ImageData
from lenstronomy.Data.psf import PSF
from lenstronomy.ImSim.differential_extinction import DifferentialExtinction
from lenstronomy.Util import util

from lenstronomy.ImSim.tracer_model import TracerModelSource


class TestTracerModel(object):
    """Test TracerModel class."""

    def setup_method(self):
        # data specifics
        sigma_bkg = 0.05  # background noise per pixel
        exp_time = 100  # exposure time (arbitrary units, flux per pixel is in units #photons/exp_time unit)
        numPix = 100  # cutout pixel size
        deltaPix = 0.05  # pixel size in arcsec (area per pixel = deltaPix**2)
        fwhm = 0.5  # full width half max of PSF

        # PSF specification

        kwargs_data = sim_util.data_configure_simple(
            numPix, deltaPix, exp_time, sigma_bkg, inverse=True
        )
        data_class = ImageData(**kwargs_data)
        kwargs_psf = {
            "psf_type": "GAUSSIAN",
            "fwhm": fwhm,
            "truncation": 5,
            "pixel_size": deltaPix,
        }
        psf_class = PSF(**kwargs_psf)

        # 'EXERNAL_SHEAR': external shear
        kwargs_sis = {"theta_E": 1.0, "center_x": 0, "center_y": 0}
        lens_model_list = ["SIS"]
        self.kwargs_lens = [kwargs_sis]
        lens_model_class = LensModel(lens_model_list=lens_model_list)
        # list of light profiles (for lens and source)
        # 'SERSIC': spherical Sersic profile
        kwargs_sersic = {
            "amp": 1.0,
            "R_sersic": 0.4,
            "n_sersic": 2,
            "center_x": 0,
            "center_y": 0,
        }
        source_light_model_list = ["SERSIC"]
        self.kwargs_source_light = [kwargs_sersic]
        source_model_class = LightModel(light_model_list=source_light_model_list)

        # Tracer model
        tracer_model = ["LINEAR"]
        self.kwargs_tracer = [{"amp": 1, "k": 2, "center_x": 0, "center_y": 0}]
        tracer_source_class = LightModel(light_model_list=tracer_model)
        kwargs_numerics = {
            "supersampling_factor": 2,
            "supersampling_convolution": False,
        }
        self.linearTracerModel = TracerModelSource(
            data_class,
            psf_class=psf_class,
            lens_model_class=lens_model_class,
            source_model_class=source_model_class,
            tracer_source_class=tracer_source_class,
            kwargs_numerics=kwargs_numerics,
            tracer_type="LINEAR",
        )
        self.logTracerModel = TracerModelSource(
            data_class,
            psf_class=psf_class,
            lens_model_class=lens_model_class,
            source_model_class=source_model_class,
            tracer_source_class=tracer_source_class,
            kwargs_numerics=kwargs_numerics,
            tracer_type="LOG",
        )

    def test_linear_tracer_model(self):
        tracer_model = self.linearTracerModel.tracer_model(
            self.kwargs_tracer,
            kwargs_lens=self.kwargs_lens,
            kwargs_source=self.kwargs_source_light,
        )

        light_unconvolved = self.linearTracerModel.source_surface_brightness(
            kwargs_source=self.kwargs_source_light,
            kwargs_lens=self.kwargs_lens,
            unconvolved=True,
        )

        light_convolved = self.linearTracerModel.source_surface_brightness(
            kwargs_source=self.kwargs_source_light,
            kwargs_lens=self.kwargs_lens,
            unconvolved=False,
        )

        source_light_num = (
            self.linearTracerModel._source_surface_brightness_analytical_numerics(
                self.kwargs_source_light, self.kwargs_lens, de_lensed=False
            )
        )
        tracer = self.linearTracerModel._tracer_model_source(
            self.kwargs_tracer, self.kwargs_lens, de_lensed=False
        )
        tracer_brightness_conv = self.linearTracerModel.ImageNumerics.re_size_convolve(
            tracer * source_light_num, unconvolved=False
        )
        tracer_model_2 = tracer_brightness_conv / light_convolved
        npt.assert_almost_equal(tracer_model_2, tracer_model, decimal=5)

    def test_log_tracer_model(self):
        """Convolve a tracer with log units with a surface brightness profile in two ways:
        (1) using the tracer_type="LOG" argument
        (2) converting units to linear ones, using tracer_type="LINEAR", and
        converting back to log units."""
        tracer = self.logTracerModel._tracer_model_source(
            self.kwargs_tracer, self.kwargs_lens, de_lensed=False
        )

        source_light_num = (
            self.logTracerModel._source_surface_brightness_analytical_numerics(
                self.kwargs_source_light, self.kwargs_lens, de_lensed=False
            )
        )

        light_unconvolved = self.linearTracerModel.source_surface_brightness(
            kwargs_source=self.kwargs_source_light,
            kwargs_lens=self.kwargs_lens,
            unconvolved=True,
        )

        light_convolved = self.linearTracerModel.source_surface_brightness(
            kwargs_source=self.kwargs_source_light,
            kwargs_lens=self.kwargs_lens,
            unconvolved=False,
        )

        # Uses tracer_type='LOG'
        tracer_model_1 = self.logTracerModel.tracer_model(
            self.kwargs_tracer,
            kwargs_lens=self.kwargs_lens,
            kwargs_source=self.kwargs_source_light,
        )

        # Converts to linear units; convolves; converts back.
        linearized_tracer = np.power(10, tracer)
        linearized_tracer_brightness_conv = (
            self.logTracerModel.ImageNumerics.re_size_convolve(
                linearized_tracer * source_light_num, unconvolved=False
            )
        )
        tracer_model_2 = np.log10(linearized_tracer_brightness_conv / light_convolved)

        npt.assert_almost_equal(tracer_model_1, tracer_model_2, decimal=5)



__author__ = "sibirrer", "gilmanda", "ajshajib"

import numpy as np
import numpy.testing as npt
import pytest

from lenstronomy.Cosmo.nfw_param import NFWParam
from lenstronomy.Cosmo.gnfw_param import GNFWParam
from astropy.cosmology import FlatLambdaCDM


class TestGNFWParam(object):
    """Tests the UnitManager class routines."""

    def setup_method(self):
        self.gnfwParam = GNFWParam(cosmo=None)
        self.nfwParam = NFWParam(cosmo=None)
        self.z = 0.5  # needed fixed redshift for the inversion function

    def test_rho0_c(self):
        c_list = [0.1, 1, 4, 10, 20]
        for c in c_list:
            rho0 = self.gnfwParam.rho0_c(c, z=self.z, gamma_in=0.8)
            c_out = self.gnfwParam.c_rho0(rho0, z=self.z, gamma_in=0.8)
            print(c, "c")
            npt.assert_almost_equal(c_out, c, decimal=3)

            nfw_rho0 = self.nfwParam.rho0_c(c, z=self.z)
            gnfw_rho0 = self.gnfwParam.rho0_c(c, z=self.z, gamma_in=1)
            npt.assert_almost_equal(nfw_rho0 / gnfw_rho0, 1.0, decimal=10)

    def test_M_r200(self):
        r200 = 200
        M200 = self.gnfwParam.M_r200(r200, z=self.z)
        M200_nfw = self.nfwParam.M_r200(r200, z=self.z)
        npt.assert_almost_equal(M200 / M200_nfw, 1, decimal=10)

    def test_rhoc_z(self):
        z = 0
        rho0_z = self.gnfwParam.rhoc_z(z=z)
        npt.assert_almost_equal(self.gnfwParam.rhoc * (1 + z) ** 3, rho0_z)

    def test_M200(self):
        M200 = self.gnfwParam.M200(rs=1, rho0=1, c=1, gamma_in=1.0)
        M200_nfw = self.nfwParam.M200(rs=1, rho0=1, c=1)
        npt.assert_almost_equal(M200 / M200_nfw, 1.0, decimal=10)

    def test_profileMain(self):
        M = 10**13.5
        z = 0.5
        gamma_in = 0.8
        r200, rho0, c, Rs = self.gnfwParam.gnfw_Mz(M, z, gamma_in)

        c_ = self.gnfwParam.c_M_z(M, z)
        r200_ = self.gnfwParam.r200_M(M, z)
        rho0_ = self.gnfwParam.rho0_c(c, z, gamma_in)
        Rs_ = r200_ / c_
        npt.assert_almost_equal(c_, c, decimal=5)
        npt.assert_almost_equal(r200_, r200, decimal=5)
        npt.assert_almost_equal(rho0_, rho0, decimal=5)
        npt.assert_almost_equal(Rs_, Rs, decimal=5)


if __name__ == "__main__":
    pytest.main()



__author__ = "sibirrer"

import numpy as np
import lenstronomy.Util.constants as const
from lenstronomy.Cosmo.cosmo_interp import CosmoInterp

__all__ = ["Background"]


class Background(object):
    """Class to compute cosmological distances."""

    def __init__(self, cosmo=None, interp=False, **kwargs_interp):
        """

        :param cosmo: instance of astropy.cosmology
        :param interp: boolean, if True, uses interpolated cosmology to evaluate specific redshifts
        :param kwargs_interp: keyword arguments of CosmoInterp specifying the interpolation interval and maximum
         redshift
        :return: Background class with instance of astropy.cosmology
        """

        if cosmo is None:
            from astropy.cosmology import default_cosmology

            cosmo = default_cosmology.get()
        if interp:
            self.cosmo = CosmoInterp(cosmo, **kwargs_interp)
        else:
            self.cosmo = cosmo

    @staticmethod
    def a_z(z):
        """Returns scale factor (a_0 = 1) for given redshift.

        :param z: redshift
        :return: scale factor
        """
        return 1.0 / (1 + z)

    def d_xy(self, z_observer, z_source):
        """

        :param z_observer: observer redshift
        :param z_source: source redshift
        :return: angular diameter distance in units of Mpc
        """
        D_xy = self.cosmo.angular_diameter_distance_z1z2(z_observer, z_source)
        return D_xy.value

    def ddt(self, z_lens, z_source):
        """Time-delay distance.

        :param z_lens: redshift of lens
        :param z_source: redshift of source
        :return: time-delay distance in units of proper Mpc
        """
        return (
            self.d_xy(0, z_lens)
            * self.d_xy(0, z_source)
            / self.d_xy(z_lens, z_source)
            * (1 + z_lens)
        )

    def T_xy(self, z_observer, z_source):
        """

        :param z_observer: observer
        :param z_source: source
        :return: transverse comoving distance in units of Mpc
        """
        D_xy = self.d_xy(z_observer, z_source)
        T_xy = D_xy * (1 + z_source)
        return T_xy

    @property
    def rho_crit(self):
        """Critical density.

        :return: value in M_sol/Mpc^3
        """
        h = self.cosmo.H(0).value / 100.0
        return 3 * h**2 / (8 * np.pi * const.G) * 10**10 * const.Mpc / const.M_sun

    def beta_double_source_plane(self, z_lens, z_source_1, z_source_2):
        """Model prediction of ratio of scaled deflection angles.

        :param z_lens: lens redshift
        :param z_source_1: source_1 redshift
        :param z_source_2: source_2 redshift
        :param cosmo: ~astropy.cosmology instance
        :return: beta
        """
        ds1 = self.cosmo.angular_diameter_distance(z=z_source_1).value
        dds1 = self.cosmo.angular_diameter_distance_z1z2(z1=z_lens, z2=z_source_1).value
        ds2 = self.cosmo.angular_diameter_distance(z=z_source_2).value
        dds2 = self.cosmo.angular_diameter_distance_z1z2(z1=z_lens, z2=z_source_2).value
        beta = dds1 / ds1 * ds2 / dds2
        return beta

    def ddt_scaling(self, z_lens, z_source_1, z_source_2):
        """Scales the time-delay distance Ddt when given for one source redshift to a
        second source redshift.

        :param z_lens: deflector redshift
        :param z_source_1: source redshift of original Ddt
        :param z_source_2: new source redshift
        :return: Ddt to z_source_2
        """
        return 1.0 / self.ddt(z_lens, z_source_1) * self.ddt(z_lens, z_source_2)



__author__ = "ajshajib"

import numpy as np

__all__ = ["MultiPlaneOrganizer"]


class MultiPlaneOrganizer(object):
    """This class organizes the lens and source planes in multi-lens plane and multi-
    source plane setting.

    In the multi-lens-plane setting with $P$ lens planes (and the last source plane
    being the $P+1$-th plane), the effective Fermat potential is defined as (eq. 9 of Shajib et al. 2020):

    .. math::
    \\phi^{\\rm eff} (\\theta) \\equiv \\sum_{i=1}^{P} \\frac{1+z_i}{1+z_{\\rm d}} \\frac{D_i D_{i+1} D_{\\rm ds}}{D_{\\rm d} D_{\\rm s}D_{i\\ i+1} } \\left[ \\frac{(\\theta_{i} - \\theta_{ i+1})^2}{2} - \\zeta_{i,i+1} \\psi_{i}(\\theta_{i})  \\right].

    Then, the $a_i$ and $b_i$ distance ratios are defined as:

    .. math::
    a_i \\equiv \\frac{D_i D_{i+1} D_{1\\ P+1}}{D_{\\rm 1} D_{P+1}D_{i\\ i+1} } = \\frac{D_i D_{i+1}}{D_{i\\ i+1} D_{\\Delta t}^{\\rm eff}}
    b_i \\equiv \\frac{D_i D_{1\\ P+1}}{D_1 D_{i\\ P+1}} = \\frac{D_i D_{P+1}}{D_{i\\ P+1} D_{\\Delta t}^{\\rm eff}}.

    This class converts the $a_i$ and $b_i$ ratios to the relevant cosmological
    distances needed for ray-tracing. However, instead of using $a_i$ and $b_i$,
    lenstronomy uses factor_a and factor_b parameters that are defined using a
    fiducial cosmology as follows:

    factor_a_i = a_i / a_i_fiducial
    factor_b_i = b_i / b_i_fiducial
    """

    def __init__(
        self,
        lens_redshift_list,
        source_redshift_list,
        sorted_lens_redshift_index,
        sorted_source_redshift_index,
        z_lens_convention,
        z_source_convention,
        cosmo,
    ):
        """Initialize the `MultiPlaneOrganizer` class.

        :param lens_redshift_list: list of lens redshifts
        :type lens_redshift_list: list
        :param source_redshift_list: list of source redshifts
        :type source_redshift_list: list
        :param sorted_lens_redshift_index: sorted index of lens redshifts
        :type sorted_lens_redshift_index: list
        :param sorted_source_redshift_index: sorted index of source redshifts
        :type sorted_source_redshift_index: list
        :param z_lens_convention: lens convention redshift
        :type z_lens_convention: float
        :param z_source_convention: source convention redshift
        :type z_source_convention: float
        :param cosmo: instance of Background class
        :type cosmo: lenstronomy.Cosmo.background.Background
        """
        self._lens_redshift_list = lens_redshift_list
        self._source_redshift_list = source_redshift_list

        self._sorted_lens_redshift_index = sorted_lens_redshift_index
        self._sorted_source_redshift_index = sorted_source_redshift_index

        self._sorted_joint_unique_redshift_list = sorted(
            list(set(list(lens_redshift_list) + list(source_redshift_list)))
        )

        self._num_lens_planes = (
            len(self._sorted_joint_unique_redshift_list) - 1
        )  # not including the last source plane
        # self._sorted_unique_lens_redshifts = sorted(list(set(
        #     lens_redshift_list)))

        self.a_coeffs_fiducial = []
        self.b_coeffs_fiducial = []
        self._D_z_list_fiducial = []
        self._D_is_list_fiducial = (
            []
        )  # distance between lens planes and the last (source) plane
        self._cosmo_bkg = cosmo

        D_s = self._cosmo_bkg.d_xy(0, z_source_convention)
        if z_lens_convention != np.min(lens_redshift_list):
            raise ValueError("z_lens_convention needs to be the first lens plane!")
        if z_source_convention != np.max(source_redshift_list):
            raise ValueError("z_source_convention needs to be the last source plane!")
        self.z_lens_convention = z_lens_convention
        self.z_source_convention = z_source_convention

        self.D_dt_eff_fiducial = (
            (1 + z_lens_convention)
            * D_s
            * self._cosmo_bkg.d_xy(0, z_lens_convention)
            / self._cosmo_bkg.d_xy(z_lens_convention, z_source_convention)
        )

        self._D_is_list_fiducial.append(
            self._cosmo_bkg.d_xy(0, self.z_source_convention)
        )

        for i in range(len(self._sorted_joint_unique_redshift_list) - 1):
            z_i = self._sorted_joint_unique_redshift_list[i]
            z_ip1 = self._sorted_joint_unique_redshift_list[i + 1]

            self._D_z_list_fiducial.append(self._cosmo_bkg.d_xy(0, z_i))
            self._D_is_list_fiducial.append(
                self._cosmo_bkg.d_xy(z_i, self.z_source_convention)
            )

            self.a_coeffs_fiducial.append(
                self._cosmo_bkg.d_xy(0, z_i)
                * self._cosmo_bkg.d_xy(0, z_ip1)
                / self._cosmo_bkg.d_xy(z_i, z_ip1)
                / self.D_dt_eff_fiducial
            )
            self.b_coeffs_fiducial.append(
                self._cosmo_bkg.d_xy(0, z_i)
                * D_s
                / self._cosmo_bkg.d_xy(z_i, z_source_convention)
                / self.D_dt_eff_fiducial
            )

        self._D_z_list_fiducial.append(
            self._cosmo_bkg.d_xy(0, self.z_source_convention)
        )

    def _extract_a_b_factors(self, kwargs_special):
        """Extracts the a and b factors from the kwargs_special dictionary.

        :param kwargs_special: dictionary of special keyword arguments
        :type kwargs_special: dict
        :return: a_factors, b_factors
        :rtype: list, list
        """
        a_factors = []
        b_factors = [1.0]

        for i in range(1, self._num_lens_planes + 1):
            a_factors.append(kwargs_special["factor_a_{}".format(i)])
        for i in range(2, self._num_lens_planes):
            b_factors.append(kwargs_special["factor_b_{}".format(i)])
        b_factors.append(a_factors[-1])

        return a_factors, b_factors

    def update_lens_T_lists(self, lens_model, kwargs_special):
        """Updates the lens model's `T_ij`, `T_ij_start`, `T_ij_stop`, and `T_z lists`.

        :param lens_model: instance of LensModel class
        :type lens_model: lenstronomy.LensModel.lens_model.LensModel
        :param kwargs_special: dictionary of special keyword arguments
        :type kwargs_special: dict
        :return: None
        :rtype: None
        """
        T_z_list, T_ij_list = self._get_lens_T_lists(kwargs_special)
        T_ij_start, T_ij_stop = self._transverse_distance_start_stop(
            0, lens_model.lens_model.z_source, kwargs_special, include_z_start=False
        )
        lens_model.lens_model.multi_plane_base.T_z_list = T_z_list
        lens_model.lens_model.multi_plane_base.T_ij_list = T_ij_list
        lens_model.lens_model.T_ij_start = T_ij_start
        lens_model.lens_model._T_ij_stop = T_ij_stop

    def update_source_mapping_T_lists(self, source_mapping_class, kwargs_special):
        """Updates the source mapping class's `T_ij_start_list` and `T_ij_end_list`.

        :param source_mapping_class: instance of SourceMapping class
        :type source_mapping_class: lenstronomy.LensModel.Solver.source_mapping.SourceMapping
        :param kwargs_special: dictionary of special keyword arguments
        :type kwargs_special: dict
        """
        T_ij_start_list, T_ij_end_list = self._get_source_T_start_end_lists(
            kwargs_special
        )
        source_mapping_class.T_ij_start_list = T_ij_start_list
        source_mapping_class.T_ij_end_list = T_ij_end_list

    def _get_element_index(self, arr, element):
        """Returns the index of an element in an array.

        :param arr: array
        :type arr: list
        :param element: element to find
        :type element: float
        :return: index of element in array
        :rtype: int
        """
        if element not in arr:
            raise ValueError("The element is not in the array!")
        index = int(np.where(np.array(arr) == element)[0][0])

        return index

    def _get_lens_T_lists(self, kwargs_special):
        """Retreive the lens model's `T_ij` and `T_z` lists for a given set of a_factors
        and b_factors.

        :param kwargs_special: dictionary of special keyword arguments
        :type kwargs_special: dict
        """
        T_ij_list = []
        T_z_list = []
        z_before = 0

        for idex in self._sorted_lens_redshift_index:
            z_lens = self._lens_redshift_list[idex]
            if z_before == z_lens:
                delta_T = 0
            else:
                # T_z = self._cosmo_bkg.T_xy(0, z_lens)
                # delta_T = self._cosmo_bkg.T_xy(z_before, z_lens)
                T_z = self._get_D_i(z_lens, kwargs_special) * (1 + z_lens)
                delta_T = self._get_D_ij(z_before, z_lens, kwargs_special) * (
                    1 + z_lens
                )

            T_ij_list.append(delta_T)
            T_z_list.append(T_z)
            z_before = z_lens

        return T_z_list, T_ij_list

    def _get_D_ij(self, z_i, z_j, kwargs_special):
        """Returns the transverse distance between two redshifts for a given set of
        a_factors and b_factors.

        :param z_i: redshift of first plane
        :type z_i: float
        :param z_j: redshift of second plane
        :type z_j: float
        :param kwargs_special: dictionary of special keyword arguments
        :type kwargs_special: dict
        :return: transverse distance between two redshifts
        :rtype: float
        """
        if z_i == 0:
            return self._get_D_i(z_j, kwargs_special)
        elif z_i == z_j:
            return 0.0
        elif z_j == self._sorted_joint_unique_redshift_list[-1]:
            ab_fiducial_index = self._get_element_index(
                self._sorted_joint_unique_redshift_list, z_i
            )
            return self._D_is_list_fiducial[ab_fiducial_index + 1]

        a_factors, b_factors = self._extract_a_b_factors(kwargs_special)
        ab_fiducial_index = self._get_element_index(
            self._sorted_joint_unique_redshift_list, z_j
        )

        b_j = b_factors[ab_fiducial_index] * self.b_coeffs_fiducial[ab_fiducial_index]
        D_j = (
            b_j
            * self._D_is_list_fiducial[ab_fiducial_index + 1]
            * self.D_dt_eff_fiducial
            / self._D_is_list_fiducial[0]
        )

        if ab_fiducial_index == 0:
            raise ValueError("The code should not come here!")
            # return D_j
        else:
            ab_fiducial_index_m1 = self._get_element_index(
                self._sorted_joint_unique_redshift_list, z_i
            )
            assert ab_fiducial_index_m1 == ab_fiducial_index - 1

            a_j = (
                a_factors[ab_fiducial_index] * self.a_coeffs_fiducial[ab_fiducial_index]
            )
            a_i = (
                a_factors[ab_fiducial_index - 1]
                * self.a_coeffs_fiducial[ab_fiducial_index - 1]
            )

            b_i = (
                b_factors[ab_fiducial_index - 1]
                * self.b_coeffs_fiducial[ab_fiducial_index - 1]
            )
            D_i = (
                b_i
                * self._D_is_list_fiducial[ab_fiducial_index]
                * self.D_dt_eff_fiducial
                / self._D_is_list_fiducial[0]
            )

            D_ij = D_j * D_i / a_i / self.D_dt_eff_fiducial

            return D_ij

    def _get_D_i(self, z_i, kwargs_special):
        """"""
        if z_i == 0.0:
            return 0.0
        elif z_i == self._sorted_joint_unique_redshift_list[-1]:
            return self._D_is_list_fiducial[0]

        a_factors, b_factors = self._extract_a_b_factors(kwargs_special)

        ab_fiducial_index = self._get_element_index(
            self._sorted_joint_unique_redshift_list, z_i
        )

        b_i = b_factors[ab_fiducial_index] * self.b_coeffs_fiducial[ab_fiducial_index]
        D_i = (
            b_i
            * self._D_is_list_fiducial[ab_fiducial_index + 1]
            * self.D_dt_eff_fiducial
            / self._D_is_list_fiducial[0]
        )

        return D_i

    def _transverse_distance_start_stop(
        self, z_start, z_stop, kwargs_special, include_z_start=False
    ):
        """Computes the transverse distance (T_ij) that is required by the ray-tracing
        between the starting redshift and the first deflector afterwards and the last
        deflector before the end of the ray-tracing.

        :param z_start: redshift of the start of the ray-tracing
        :param z_stop: stop of ray-tracing
        :param include_z_start: boolean, if True includes the computation of the
            starting position if the first deflector is at z_start
        :return: T_ij_start, T_ij_end
        """
        z_lens_last = z_start
        first_deflector = True
        T_ij_start = None
        for i, idex in enumerate(self._sorted_lens_redshift_index):
            z_lens = self._lens_redshift_list[idex]
            if (
                self._start_condition(include_z_start, z_lens, z_start)
                and z_lens <= z_stop
            ):
                if first_deflector is True:
                    T_ij_start = self._get_D_ij(z_start, z_lens, kwargs_special) * (
                        1 + z_lens
                    )
                    first_deflector = False
                z_lens_last = z_lens
        T_ij_end = self._get_D_ij(z_lens_last, z_stop, kwargs_special) * (1 + z_stop)
        return T_ij_start, T_ij_end

    def _get_source_T_start_end_lists(self, kwargs_special, include_z_start=False):
        """"""
        # self._sorted_source_redshift_index
        z_start = 0
        T_ij_start_list = []
        T_ij_end_list = []

        for i, index_source in enumerate(self._sorted_source_redshift_index):
            z_stop = self._source_redshift_list[index_source]
            T_ij_start, T_ij_end = self._transverse_distance_start_stop(
                z_start, z_stop, kwargs_special, include_z_start=False
            )
            T_ij_start_list.append(T_ij_start)
            T_ij_end_list.append(T_ij_end)
            z_start = z_stop

        return T_ij_start_list, T_ij_end_list

    @staticmethod
    def _start_condition(inclusive, z_lens, z_start):
        """Boolean condition if the starting redshift is met.

        :param inclusive: boolean, if True selects z_lens including z_start, else only
            selects z_lens > z_start
        :param z_lens: deflector redshift
        :param z_start: starting redshift (lowest redshift)
        :return: boolean of condition
        """

        if inclusive:
            return z_lens >= z_start
        else:
            return z_lens > z_start



import numpy as np
import pytest
from lenstronomy.LensModel.Microlensing.adaptive_boundary_mesh import splitting_centers
from lenstronomy.LensModel.Microlensing.adaptive_boundary_mesh import loop_information
from lenstronomy.LensModel.Microlensing.adaptive_boundary_mesh import within_distance
from lenstronomy.LensModel.Microlensing.adaptive_boundary_mesh import (
    adaptive_boundary_mesh,
)


class Test_adaptive_boundary_mesh:

    def test_splitting_centers(self):
        """Function to test the splitting centers functionality."""

        center = np.array([[0, 0]])
        x = center[:, 0]
        y = center[:, 1]
        side_length = 1
        n_p = 2
        expected_x = np.array([-0.25, 0.25, -0.25, 0.25])
        expected_y = np.array([-0.25, -0.25, 0.25, 0.25])

        new_x, new_y, new_side_length = splitting_centers(x, y, side_length, n_p)
        assert np.all(new_x == expected_x)
        assert np.all(new_y == expected_y)
        assert new_side_length == pytest.approx(0.5, rel=1e-7)

    def test_loop_information(self):
        """Function to test the loop information functionality."""

        eta = 2
        beta_0 = 50
        beta_s = 10
        expected_iterations = 4
        expected_final_eta = 1.25

        number_of_iterations, final_eta = loop_information(eta, beta_0, beta_s)
        assert number_of_iterations == expected_iterations
        assert final_eta == pytest.approx(expected_final_eta, rel=1e-7)

    def test_within_distance(self):

        center_points_within = np.array([[2, 2], [3, 2], [4, 3]])
        center_points_not_within = np.array([[1, 1], [1, 2], [5, 5]])
        test_point = (3, 2)
        delta_beta = 2
        beta_s = 3

        # Test points within the threshold
        distances_within = within_distance(
            center_points_within[:, 0],
            center_points_within[:, 1],
            test_point,
            delta_beta,
            beta_s,
        )
        assert np.all(distances_within)

        # Test points not within the threshold
        distances_not_within = within_distance(
            center_points_not_within[:, 0],
            center_points_not_within[:, 1],
            test_point,
            delta_beta,
            beta_s,
        )
        assert not np.any(distances_not_within)

    def test_ABM(self):
        """Test function for ABM.

        It sets up the initial parameters and checks if the function returns the
        expected results.
        """

        source_position = (0, 0)
        L = 0.0004
        beta_0 = 0.0016
        beta_s = 1.16e-5
        n_p = 5
        eta = 0.7 * n_p
        number_of_iterations = 5
        final_eta = 3.23
        kwargs_lens = [
            {"theta_E": 0.0001, "center_x": 0.000025, "center_y": 0.00001666666}
        ]

        (
            side_length,
            total_number_of_rays_shot,
            image_subset_centers_x,
            image_subset_centers_y,
        ) = adaptive_boundary_mesh(
            source_position,
            L,
            beta_0,
            beta_s,
            n_p,
            eta,
            number_of_iterations,
            final_eta,
            kwargs_lens,
        )

        expected_number_of_subset_centers = 3576
        expected_side_length = 6.45818e-7
        expected_total_number_of_rays_shot = 3724

        assert len(image_subset_centers_x) == len(image_subset_centers_y)
        assert len(image_subset_centers_x) == expected_number_of_subset_centers
        assert side_length == pytest.approx(expected_side_length, rel=1e-12)
        assert total_number_of_rays_shot == expected_total_number_of_rays_shot



from lenstronomy.Data.imaging_data import ImageData
from lenstronomy.Data.psf import PSF
from lenstronomy.LensModel.lens_model import LensModel
from lenstronomy.LightModel.light_model import LightModel
from lenstronomy.PointSource.point_source import PointSource
from lenstronomy.ImSim.differential_extinction import DifferentialExtinction
from lenstronomy.ImSim.image_linear_solve import ImageLinearFit
from lenstronomy.ImSim.tracer_model import TracerModelSource

from lenstronomy.Util.package_util import exporter

export, __all__ = exporter()


@export
def create_class_instances(
    lens_model_list=None,
    z_lens=None,
    z_source=None,
    z_source_convention=None,
    lens_redshift_list=None,
    kwargs_interp=None,
    multi_plane=False,
    distance_ratio_sampling=False,
    observed_convention_index=None,
    source_light_model_list=None,
    lens_light_model_list=None,
    point_source_model_list=None,
    point_source_redshift_list=None,
    fixed_magnification_list=None,
    flux_from_point_source_list=None,
    point_source_frame_list=None,
    additional_images_list=None,
    kwargs_lens_eqn_solver=None,
    source_deflection_scaling_list=None,
    source_redshift_list=None,
    cosmo=None,
    index_lens_model_list=None,
    index_source_light_model_list=None,
    index_lens_light_model_list=None,
    index_point_source_model_list=None,
    optical_depth_model_list=None,
    index_optical_depth_model_list=None,
    band_index=0,
    tau0_index_list=None,
    all_models=False,
    point_source_magnification_limit=None,
    surface_brightness_smoothing=0.001,
    sersic_major_axis=None,
    tabulated_deflection_angles=None,
    decouple_multi_plane=False,
    kwargs_multiplane_model=None,
    kwargs_multiplane_model_point_source=None,
    tracer_source_model_list=None,
    tracer_source_band=0,
    tracer_partition=None,
    tracer_type="LINEAR",
):
    """

    :param lens_model_list: list of strings indicating the type of lens models
    :param z_lens: redshift of the deflector (for single lens plane mode, but only relevant when computing physical quantities)
    :param z_source: redshift of source (for single source plane mode, or for multiple source planes the redshift of the point source). In regard to this redshift the reduced deflection angles are defined in the lens model.
    :param z_source_convention: float, redshift of a source to define the reduced deflection angles of the lens models.
     If None, 'z_source' is used.
    :param lens_redshift_list:
    :param multi_plane: bool, if True, computes the lensing quantities in multi-plane mode
    :param distance_ratio_sampling: bool, if True, samples the distance ratios in multi-lens-plane
    :param kwargs_interp: interpolation keyword arguments specifying the numerics.
     See description in the Interpolate() class. Only applicable for 'INTERPOL' and 'INTERPOL_SCALED' models.
    :param observed_convention_index:
    :param source_light_model_list:
    :param lens_light_model_list:
    :param point_source_model_list:
    :param fixed_magnification_list:
    :param flux_from_point_source_list: list of bools (optional), if set, will only return image positions
         (for imaging modeling) for the subset of the point source lists that =True. This option enables to model
    :param point_source_frame_list: list of lists mirroring the structure of the image positions.
     Integers correspond to the i'th list entry of index_lens_model_list indicating in which frame/band the image is
     appearing
    :param additional_images_list:
    :param kwargs_lens_eqn_solver: keyword arguments specifying the numerical settings for the lens equation solver
         see LensEquationSolver() class for details
    :param source_deflection_scaling_list: List of floats for each source ligth model (optional, and only applicable
     for single-plane lensing. The factors re-scale the reduced deflection angles described from the lens model.
     =1 means identical source position as without this option. This option enables multiple source planes.
     The geometric difference between the different source planes needs to be pre-computed and is cosmology dependent.
    :param source_redshift_list:
    :param cosmo: astropy.cosmology instance
    :param index_lens_model_list:
    :param index_source_light_model_list:
    :param index_lens_light_model_list: optional, list of list of all model indexes for each modeled band
    :param index_point_source_model_list: optional, list of list of all model indexes for each modeled band
    :param optical_depth_model_list: list of strings indicating the optical depth model to compute (differential) extinctions from the source
    :param index_optical_depth_model_list:
    :param band_index: int, index of band to consider. Has an effect if only partial models are considered for a specific band
    :param tau0_index_list: list of integers of the specific extinction scaling parameter tau0 for each band
    :param all_models: bool, if True, will make class instances of all models ignoring potential keywords that are excluding specific models as indicated.
    :param point_source_magnification_limit: float >0 or None, if set and additional images are computed, then it will cut the point sources computed to the limiting (absolute) magnification
    :param surface_brightness_smoothing: float, smoothing scale of light profile (minimal distance to the center of a profile)
     this can help to avoid inaccuracies in the very center of a cuspy light profile
    :param sersic_major_axis: boolean or None, if True, uses the semi-major axis as the definition of the Sersic
     half-light radius, if False, uses the product average of semi-major and semi-minor axis. If None, uses the
     convention in the lenstronomy yaml setting (which by default is =False)
    :param tabulated_deflection_angles: a user-specified class with a call method that returns deflection angles given
     (x, y) coordinates on the sky. This class gets passed to the lens model class TabulatedDeflections
    :param decouple_multi_plane: bool; if True, creates an instance of MultiPlaneDecoupled
    :param kwargs_multiplane_model: keyword arguments used to create an instance of MultiPlaneDecoupled if decouple_multi_plane is True
    :param kwargs_multiplane_model_point_source: keyword arguments used to create an option MultiPlaneDecoupled class for the lensed point source to be treated separately from the rest of the imaging data
    :param tracer_source_model_list: list of tracer source models (not used in this function)
    :param tracer_source_band: integer, list index of source surface brightness band to apply tracer model to
    :param tracer_partition: in case of tracer models for specific sub-parts of the surface brightness model
     [[list of light profiles, list of tracer profiles], [list of light profiles, list of tracer profiles], [...], ...]
    :type tracer_partition: None or list
    :param tracer_type: 'LINEAR' or 'LOG', to determine how tracers are summed between components
    :type tracer_type: string
    :param point_source_redshift_list: list of redshifts of point sources
         (default None, i.e. all point sources at the same redshift following the source convention)
    :return: lens_model_class, source_model_class, lens_light_model_class, point_source_class, extinction_class
    """
    if lens_model_list is None:
        lens_model_list = []
    if lens_light_model_list is None:
        lens_light_model_list = []
    if source_light_model_list is None:
        source_light_model_list = []
    if point_source_model_list is None:
        point_source_model_list = []

    if index_lens_model_list is None or all_models is True:
        lens_model_list_i = lens_model_list
        lens_redshift_list_i = lens_redshift_list
        observed_convention_index_i = observed_convention_index
    else:
        lens_model_list_i = [
            lens_model_list[k] for k in index_lens_model_list[band_index]
        ]
        if lens_redshift_list is not None:
            lens_redshift_list_i = [
                lens_redshift_list[k] for k in index_lens_model_list[band_index]
            ]
        else:
            lens_redshift_list_i = lens_redshift_list
        if observed_convention_index is not None:
            counter = 0
            observed_convention_index_i = []
            for k in index_lens_model_list[band_index]:
                if k in observed_convention_index:
                    observed_convention_index_i.append(counter)
                counter += 1
        else:
            observed_convention_index_i = observed_convention_index

    lens_model_class = LensModel(
        lens_model_list=lens_model_list_i,
        z_lens=z_lens,
        z_source=z_source,
        z_source_convention=z_source_convention,
        lens_redshift_list=lens_redshift_list_i,
        multi_plane=multi_plane,
        cosmo=cosmo,
        distance_ratio_sampling=distance_ratio_sampling,
        observed_convention_index=observed_convention_index_i,
        kwargs_interp=kwargs_interp,
        numerical_alpha_class=tabulated_deflection_angles,
        decouple_multi_plane=decouple_multi_plane,
        kwargs_multiplane_model=kwargs_multiplane_model,
    )

    if kwargs_multiplane_model_point_source is not None:
        lens_model_class_point_source = LensModel(
            lens_model_list=lens_model_list,
            z_lens=z_lens,
            z_source=z_source,
            z_source_convention=z_source_convention,
            lens_redshift_list=lens_redshift_list,
            multi_plane=multi_plane,
            cosmo=cosmo,
            observed_convention_index=observed_convention_index,
            kwargs_interp=kwargs_interp,
            numerical_alpha_class=tabulated_deflection_angles,
            decouple_multi_plane=decouple_multi_plane,
            kwargs_multiplane_model=kwargs_multiplane_model_point_source,
        )
    else:
        lens_model_class_point_source = lens_model_class

    if index_source_light_model_list is None or all_models is True:
        source_light_model_list_i = source_light_model_list
        source_deflection_scaling_list_i = source_deflection_scaling_list
        source_redshift_list_i = source_redshift_list
    else:
        source_light_model_list_i = [
            source_light_model_list[k]
            for k in index_source_light_model_list[band_index]
        ]
        if source_deflection_scaling_list is None:
            source_deflection_scaling_list_i = source_deflection_scaling_list
        else:
            source_deflection_scaling_list_i = [
                source_deflection_scaling_list[k]
                for k in index_source_light_model_list[band_index]
            ]
        if source_redshift_list is None:
            source_redshift_list_i = source_redshift_list
        else:
            source_redshift_list_i = [
                source_redshift_list[k]
                for k in index_source_light_model_list[band_index]
            ]
    source_model_class = LightModel(
        light_model_list=source_light_model_list_i,
        deflection_scaling_list=source_deflection_scaling_list_i,
        source_redshift_list=source_redshift_list_i,
        smoothing=surface_brightness_smoothing,
        sersic_major_axis=sersic_major_axis,
    )

    if index_lens_light_model_list is None or all_models is True:
        lens_light_model_list_i = lens_light_model_list
    else:
        lens_light_model_list_i = [
            lens_light_model_list[k] for k in index_lens_light_model_list[band_index]
        ]
    lens_light_model_class = LightModel(
        light_model_list=lens_light_model_list_i,
        smoothing=surface_brightness_smoothing,
        sersic_major_axis=sersic_major_axis,
    )

    point_source_model_list_i = point_source_model_list
    fixed_magnification_list_i = fixed_magnification_list
    additional_images_list_i = additional_images_list
    point_source_frame_list_i = point_source_frame_list

    if index_point_source_model_list is not None and not all_models:
        point_source_model_list_i = [
            point_source_model_list[k]
            for k in index_point_source_model_list[band_index]
        ]
        if fixed_magnification_list is not None:
            fixed_magnification_list_i = [
                fixed_magnification_list[k]
                for k in index_point_source_model_list[band_index]
            ]
        if additional_images_list is not None:
            additional_images_list_i = [
                additional_images_list[k]
                for k in index_point_source_model_list[band_index]
            ]
        if point_source_frame_list is not None:
            point_source_frame_list_i = [
                point_source_frame_list[k]
                for k in index_point_source_model_list[band_index]
            ]
    point_source_class = PointSource(
        point_source_type_list=point_source_model_list_i,
        lens_model=lens_model_class_point_source,
        fixed_magnification_list=fixed_magnification_list_i,
        flux_from_point_source_list=flux_from_point_source_list,
        additional_images_list=additional_images_list_i,
        magnification_limit=point_source_magnification_limit,
        kwargs_lens_eqn_solver=kwargs_lens_eqn_solver,
        point_source_frame_list=point_source_frame_list_i,
        index_lens_model_list=index_lens_model_list,
        redshift_list=point_source_redshift_list,
    )
    if tau0_index_list is None:
        tau0_index = 0
    else:
        tau0_index = tau0_index_list[band_index]
    if index_optical_depth_model_list is not None:
        optical_depth_model_list_i = [
            optical_depth_model_list[k]
            for k in index_optical_depth_model_list[band_index]
        ]
    else:
        optical_depth_model_list_i = optical_depth_model_list
    extinction_class = DifferentialExtinction(
        optical_depth_model=optical_depth_model_list_i, tau0_index=tau0_index
    )
    return (
        lens_model_class,
        source_model_class,
        lens_light_model_class,
        point_source_class,
        extinction_class,
    )


@export
def create_image_model(
    kwargs_data, kwargs_psf, kwargs_numerics, kwargs_model, image_likelihood_mask=None
):
    """

    :param kwargs_data: ImageData keyword arguments
    :param kwargs_psf: PSF keyword arguments
    :param kwargs_numerics: numerics keyword arguments for Numerics() class
    :param kwargs_model: model keyword arguments
    :param image_likelihood_mask: image likelihood mask
     (same size as image_data with 1 indicating being evaluated and 0 being left out)
    :return: ImageLinearFit() instance
    """
    data_class = ImageData(**kwargs_data)
    psf_class = PSF(**kwargs_psf)
    (
        lens_model_class,
        source_model_class,
        lens_light_model_class,
        point_source_class,
        extinction_class,
    ) = create_class_instances(**kwargs_model)
    imageModel = ImageLinearFit(
        data_class,
        psf_class,
        lens_model_class,
        source_model_class,
        lens_light_model_class,
        point_source_class,
        extinction_class,
        kwargs_numerics,
        likelihood_mask=image_likelihood_mask,
    )
    return imageModel


@export
def create_im_sim(
    multi_band_list,
    multi_band_type,
    kwargs_model,
    bands_compute=None,
    image_likelihood_mask_list=None,
    band_index=0,
    kwargs_pixelbased=None,
    linear_solver=True,
):
    """


    :param multi_band_list: list of [[kwargs_data, kwargs_psf, kwargs_numerics], [], ..]
    :param multi_band_type: string, option when having multiple imaging data sets modelled simultaneously. Options are:
     - 'multi-linear': linear amplitudes are inferred on single data set
     - 'linear-joint': linear amplitudes ae jointly inferred
     - 'single-band': single band
    :param kwargs_model: model keyword arguments
    :param bands_compute: (optional), bool list to indicate which band to be included in the modeling
    :param image_likelihood_mask_list: list of image likelihood mask
     (same size as image_data with 1 indicating being evaluated and 0 being left out)
    :param band_index: integer, index of the imaging band to model (only applied when using 'single-band' as option)
    :param kwargs_pixelbased: keyword arguments with various settings related to the pixel-based solver (see SLITronomy documentation)
    :param linear_solver: bool, if True (default) fixes the linear amplitude parameters 'amp' (avoid sampling) such
     that they get overwritten by the linear solver solution.
    :return: MultiBand class instance
    """
    if linear_solver is False and multi_band_type not in [
        "single-band",
        "multi-linear",
    ]:
        raise ValueError(
            'setting "linear_solver" to False is only supported in "single-band" mode '
            'or if "multi-linear" model has only one band.'
        )

    if multi_band_type == "multi-linear":
        from lenstronomy.ImSim.MultiBand.multi_linear import MultiLinear

        multiband = MultiLinear(
            multi_band_list,
            kwargs_model,
            compute_bool=bands_compute,
            likelihood_mask_list=image_likelihood_mask_list,
            linear_solver=linear_solver,
        )
    elif multi_band_type == "joint-linear":
        from lenstronomy.ImSim.MultiBand.joint_linear import JointLinear

        multiband = JointLinear(
            multi_band_list,
            kwargs_model,
            compute_bool=bands_compute,
            likelihood_mask_list=image_likelihood_mask_list,
        )
    elif multi_band_type == "single-band":
        from lenstronomy.ImSim.MultiBand.single_band_multi_model import (
            SingleBandMultiModel,
        )

        multiband = SingleBandMultiModel(
            multi_band_list,
            kwargs_model,
            likelihood_mask_list=image_likelihood_mask_list,
            band_index=band_index,
            kwargs_pixelbased=kwargs_pixelbased,
            linear_solver=linear_solver,
        )
    else:
        raise ValueError("type %s is not supported!" % multi_band_type)
    return multiband


def create_tracer_model(tracer_data, kwargs_model, tracer_likelihood_mask=None):
    """

    :param tracer_data:
    :param kwargs_model:
    :param tracer_likelihood_mask:

    :return:
    """
    tracer_source_band = kwargs_model.get("tracer_source_band", 0)
    tracer_source_class = LightModel(
        light_model_list=kwargs_model.get("tracer_source_model_list", [])
    )
    tracer_partition = kwargs_model.get("tracer_partition", None)
    kwargs_data, kwargs_psf, kwargs_numerics = tracer_data
    (
        lens_model_class,
        source_model_class,
        lens_light_model_class,
        point_source_class,
        extinction_class,
    ) = create_class_instances(band_index=tracer_source_band, **kwargs_model)
    data_class = ImageData(**kwargs_data)
    psf_class = PSF(**kwargs_psf)
    tracer_model = TracerModelSource(
        data_class,
        psf_class=psf_class,
        lens_model_class=lens_model_class,
        source_model_class=source_model_class,
        lens_light_model_class=lens_light_model_class,
        point_source_class=point_source_class,
        extinction_class=extinction_class,
        tracer_source_class=tracer_source_class,
        kwargs_numerics=kwargs_numerics,
        likelihood_mask=tracer_likelihood_mask,
        psf_error_map_bool_list=None,
        kwargs_pixelbased=None,
        tracer_partition=tracer_partition,
    )
    return tracer_model



__author__ = "sibirrer"


from lenstronomy.LensModel.Profiles.nfw import NFW
from lenstronomy.LensModel.Profiles.gnfw import GNFW

import numpy as np
import numpy.testing as npt
import pytest


class TestGNFW(object):
    """This class tests the generalized NFW profile."""

    def setup_method(self):
        self.nfw = NFW()
        self.gnfw = GNFW()
        self.gnfw_trapezoidal = GNFW(trapezoidal_integration=True)

    def test_function(self):
        """Tests `GNFW.function()`"""
        x = 1.5
        y = 2
        Rs = 5.0
        rho0 = 0.1
        gamma_in = 1

        alpha_Rs_nfw = self.nfw.rho02alpha(rho0, Rs)
        values_nfw = self.nfw.function(x, y, Rs, alpha_Rs_nfw)

        alpha_Rs = self.gnfw.rho02alpha(rho0, Rs, gamma_in)
        values_gnfw = self.gnfw.function(x, y, Rs, alpha_Rs, gamma_in)
        npt.assert_almost_equal(values_nfw, values_gnfw, decimal=3)

        values_gnfw_trapezoidal = self.gnfw_trapezoidal.function(
            x, y, Rs, alpha_Rs, gamma_in
        )
        npt.assert_almost_equal(values_nfw, values_gnfw_trapezoidal, decimal=4)

        # test for array of values
        x = np.linspace(0.5, 10, 10)
        y = np.ones_like(x)

        values_nfw = self.nfw.function(x, y, Rs, alpha_Rs_nfw)
        values_gnfw = self.gnfw.function(x, y, Rs, alpha_Rs, gamma_in)
        npt.assert_almost_equal(values_nfw, values_gnfw, decimal=4)

        # test for really small Rs
        Rs = 0.00000001
        alpha_Rs_nfw = self.nfw.rho02alpha(rho0, Rs)
        alpha_Rs = self.gnfw.rho02alpha(rho0, Rs, gamma_in)
        values_nfw = self.nfw.function(x, y, Rs, alpha_Rs_nfw)
        values_gnfw = self.gnfw.function(x, y, Rs, alpha_Rs, gamma_in)
        npt.assert_almost_equal(values_nfw, values_gnfw, decimal=10)

    def test_derivatives(self):
        """Tests `GNFW.derivatives()`"""
        x = np.array([1.5])
        y = np.array([2])
        Rs = 5.0
        rho0 = 0.1
        gamma_in = 1

        alpha_Rs_nfw = self.nfw.rho02alpha(rho0, Rs)
        f_x_nfw, f_y_nfw = self.nfw.derivatives(x, y, Rs, alpha_Rs_nfw)

        alpha_Rs = self.gnfw.rho02alpha(rho0, Rs, gamma_in)
        f_x_gnfw, f_y_gnfw = self.gnfw.derivatives(x, y, Rs, alpha_Rs, gamma_in)

        npt.assert_almost_equal(f_x_nfw, f_x_gnfw, decimal=8)
        npt.assert_almost_equal(f_y_nfw, f_y_gnfw, decimal=8)

        f_x_gnfwt, f_y_gnfwt = self.gnfw_trapezoidal.derivatives(
            x, y, Rs, alpha_Rs, gamma_in
        )
        npt.assert_almost_equal(f_x_nfw, f_x_gnfwt, decimal=4)
        npt.assert_almost_equal(f_y_nfw, f_y_gnfwt, decimal=4)

        # test for really small Rs
        Rs = 0.00000001
        alpha_Rs_nfw = self.nfw.rho02alpha(rho0, Rs)
        alpha_Rs = self.gnfw.rho02alpha(rho0, Rs, gamma_in)
        f_x_nfw, f_y_nfw = self.nfw.derivatives(x, y, Rs, alpha_Rs_nfw)
        f_x_gnfw, f_y_gnfw = self.gnfw.derivatives(x, y, Rs, alpha_Rs, gamma_in)
        npt.assert_almost_equal(f_x_nfw, f_x_gnfw, decimal=5)
        npt.assert_almost_equal(f_y_nfw, f_y_gnfw, decimal=5)

    def test_hessian(self):
        """Tests `GNFW.hessian()`"""
        x = np.linspace(0.5, 10, 10)
        y = x * 0.0
        Rs = 5.0
        rho0 = 0.1
        gamma_in = 1

        alpha_Rs_nfw = self.nfw.rho02alpha(rho0, Rs)
        f_xx_nfw, f_xy_nfw, _, f_yy_nfw = self.nfw.hessian(x, y, Rs, alpha_Rs_nfw)

        alpha_Rs = self.gnfw.rho02alpha(rho0, Rs, gamma_in)
        f_xx_gnfw, f_xy_gnfw, _, f_yy_gnfw = self.gnfw.hessian(
            x, y, Rs, alpha_Rs, gamma_in
        )

        npt.assert_almost_equal(f_xx_nfw, f_xx_gnfw, decimal=8)
        npt.assert_almost_equal(f_yy_nfw, f_yy_gnfw, decimal=8)
        npt.assert_almost_equal(f_xy_nfw, f_xy_gnfw, decimal=8)

        f_xx_gnfwt, f_xy_gnfwt, _, f_yy_gnfwt = self.gnfw_trapezoidal.hessian(
            x, y, Rs, alpha_Rs, gamma_in
        )
        npt.assert_almost_equal(f_xx_nfw, f_xx_gnfwt, decimal=4)
        npt.assert_almost_equal(f_yy_nfw, f_yy_gnfwt, decimal=4)
        npt.assert_almost_equal(f_xy_nfw, f_xy_gnfwt, decimal=4)

        # test for really small Rs
        Rs = 0.00000001
        alpha_Rs_nfw = self.nfw.rho02alpha(rho0, Rs)
        alpha_Rs = self.gnfw.rho02alpha(rho0, Rs, gamma_in)
        f_xx_nfw, f_xy_nfw, _, f_yy_nfw = self.nfw.hessian(x, y, Rs, alpha_Rs_nfw)
        f_xx_gnfw, f_xy_gnfw, _, f_yy_gnfw = self.gnfw.hessian(
            x, y, Rs, alpha_Rs, gamma_in
        )

        npt.assert_almost_equal(f_xx_nfw, f_xx_gnfw, decimal=2)
        npt.assert_almost_equal(f_yy_nfw, f_yy_gnfw, decimal=2)
        npt.assert_almost_equal(f_xy_nfw, f_xy_gnfw, decimal=2)

    def test_density(self):
        """Tests `GNFW.density()`"""
        R = 2
        Rs = 5.0
        rho0 = 0.1
        gamma_in = 1

        density_nfw = self.nfw.density(R, Rs, rho0)
        density_gnfw = self.gnfw.density(R, Rs, rho0, gamma_in)
        density_gnfwt = self.gnfw_trapezoidal.density(R, Rs, rho0, gamma_in)

        npt.assert_almost_equal(density_nfw, density_gnfw, decimal=10)
        npt.assert_almost_equal(density_nfw, density_gnfwt, decimal=6)

    def test_density_lens(self):
        """Tests `GNFW.density_lens()`"""
        R = 2
        Rs = 5.0
        rho0 = 0.1
        gamma_in = 1

        alpha_Rs_nfw = self.nfw.rho02alpha(rho0, Rs)
        density_nfw = self.nfw.density_lens(R, Rs, alpha_Rs_nfw)

        alpha_Rs = self.gnfw.rho02alpha(rho0, Rs, gamma_in)
        density_gnfw = self.gnfw.density_lens(R, Rs, alpha_Rs, gamma_in)
        density_gnfwt = self.gnfw_trapezoidal.density_lens(R, Rs, alpha_Rs, gamma_in)

        npt.assert_almost_equal(density_nfw, density_gnfw, decimal=9)
        npt.assert_almost_equal(density_nfw, density_gnfwt, decimal=6)

    def test_density_2d(self):
        """Tests `GNFW.density_2d_lens()`"""
        x = np.array([1.5])
        y = np.array([2])
        Rs = 5.0
        rho0 = 0.1
        gamma_in = 1

        kappa_nfw = self.nfw.density_2d(x, y, Rs, rho0)
        kappa_gnfw = self.gnfw.density_2d(x, y, Rs, rho0, gamma_in)
        kappa_gnfwt = self.gnfw_trapezoidal.density_2d(x, y, Rs, rho0, gamma_in)
        npt.assert_almost_equal(kappa_nfw, kappa_gnfw, decimal=10)
        npt.assert_almost_equal(kappa_nfw, kappa_gnfwt, decimal=5)

    def test_mass_3d(self):
        """Tests `GNFW.mass_3d()`"""
        r = 2
        Rs = 1.0
        rho0 = 1
        gamma_in = 1

        mass_3d_nfw = self.nfw.mass_3d(r, Rs, rho0)
        mass_3d_gnfw = self.gnfw.mass_3d(r, Rs, rho0, gamma_in)

        npt.assert_almost_equal(mass_3d_nfw, mass_3d_gnfw, decimal=10)

    def test_mass_3d_lens(self):
        """Tests `GNFW.mass_3d_lens()`"""
        r = 2
        Rs = 5.0
        rho0 = 0.1
        gamma_in = 1

        alpha_Rs_nfw = self.nfw.rho02alpha(rho0, Rs)
        mass_3d_nfw = self.nfw.mass_3d_lens(r, Rs, alpha_Rs_nfw)

        alpha_Rs = self.gnfw.rho02alpha(rho0, Rs, gamma_in)
        mass_3d_gnfw = self.gnfw.mass_3d_lens(r, Rs, alpha_Rs, gamma_in)

        npt.assert_almost_equal(mass_3d_nfw, mass_3d_gnfw, decimal=8)


if __name__ == "__main__":
    pytest.main()



import numpy as np
import lenstronomy.Util.param_util as param_util

from lenstronomy.Util.package_util import exporter

export, __all__ = exporter()


@export
class Linear(object):
    """Class for two-dimensional Linear light profile.

    profile name in LightModel module: 'LINEAR'
    """

    def __init__(self):
        self.param_names = ["amp", "k", "center_x", "center_y"]
        self.lower_limit_default = {
            "k": -1000,
            "amp": 0,
            "center_x": -100,
            "center_y": -100,
        }
        self.upper_limit_default = {
            "k": 1000,
            "amp": 100,
            "center_x": 100,
            "center_y": 100,
        }

    def function(self, x, y, amp, k, center_x=0, center_y=0):
        """Surface brightness per angular unit.

        :param x: coordinate on the sky
        :param y: coordinate on the sky
        :param k: slope value for the linear profile
        :param amp: the intercept value before time amp
        :param center_x: center of profile
        :param center_y: center of profile
        :return: surface brightness at (x, y)
        """
        dis = np.sqrt((x - center_x) ** 2 + (y - center_y) ** 2)
        f = amp * (k / amp * dis + 1)
        f[f < 0] = 0
        return f

    def total_flux(self, amp, k, center_x=0, center_y=0):
        """Integrated flux of the profile.

        :param k: slope value for the linear profile
        :param amp: the intercept value before time amp
        :param center_x: center of profile
        :param center_y: center of profile
        :return: total flux
        """
        return 1 / 3 * (np.pi * (k / abs(amp)) ** 2 * k)


@export
class LinearEllipse(object):
    """Class for Linear light profile with ellipticity.

    profile name in LightModel module: 'LINEAR_ELLIPSE'
    """

    param_names = ["amp", "k", "e1", "e2", "center_x", "center_y"]
    lower_limit_default = {
        "k": -1000,
        "amp": 0,
        "e1": -0.5,
        "e2": -0.5,
        "center_x": -100,
        "center_y": -100,
    }
    upper_limit_default = {
        "k": 1000,
        "amp": 100,
        "e1": -0.5,
        "e2": -0.5,
        "center_x": 100,
        "center_y": 100,
    }

    def __init__(self):
        self.linear = Linear()

    def function(self, x, y, amp, k, e1, e2, center_x=0, center_y=0):
        """

        :param x: coordinate on the sky
        :param y: coordinate on the sky
        :param k: slope value for the linear profile
        :param amp: the intercept value before time amp
        :param e1: eccentricity modulus
        :param e2: eccentricity modulus
        :param center_x: center of profile
        :param center_y: center of profile
        :return: surface brightness at (x, y)
        """
        x_, y_ = param_util.transform_e1e2_product_average(
            x, y, e1, e2, center_x, center_y
        )
        return self.linear.function(x_, y_, amp, k, center_x=0, center_y=0)

    def total_flux(self, amp, k, e1=None, e2=None, center_x=None, center_y=None):
        """Total integrated flux of profile.

        :param k: slope value for the linear profile
        :param amp: the intercept value before time amp
        :param e1: eccentricity modulus
        :param e2: eccentricity modulus
        :param center_x: center of profile
        :param center_y: center of profile
        :return: total flux
        """
        return self.linear.total_flux(amp, k, center_x, center_y)



__author__ = "ajshajib"

import numpy as np
import numpy.testing as npt
from lenstronomy.ImSim.image2source_mapping import Image2SourceMapping
from lenstronomy.ImSim.multiplane_organizer import MultiPlaneOrganizer
from lenstronomy.Cosmo.background import Background
from astropy.cosmology import FlatLambdaCDM
from lenstronomy.LensModel.lens_model import LensModel
from lenstronomy.LightModel.light_model import LightModel
import pytest
import unittest
import copy


class TestMultiPlaneOrganizer(object):
    """"""

    def setup_method(self, method):
        """"""
        self.lens_redshift_list = [0.2, 0.4, 0.6, 1.0]
        self.source_redshift_list = [0.6, 1.0, 1.0, 1.2]
        self.cosmo = FlatLambdaCDM(H0=75, Om0=0.4)

        self.lens_model = LensModel(
            ["SIS", "SIS", "SIS", "BLANK_PLANE"],
            z_source=1.2,
            lens_redshift_list=self.lens_redshift_list,
            multi_plane=True,
            z_source_convention=1.2,
            distance_ratio_sampling=True,
            cosmo=self.cosmo,
        )

        self.lens_model_fiducial = LensModel(
            ["SIS", "SIS", "SIS", "SIS"],
            z_source=1.2,
            lens_redshift_list=self.lens_redshift_list,
            multi_plane=True,
            z_source_convention=1.2,
            distance_ratio_sampling=False,
            cosmo=self.cosmo,
        )

        self.source_model = LightModel(
            ["GAUSSIAN", "GAUSSIAN", "GAUSSIAN", "GAUSSIAN"],
            source_redshift_list=self.source_redshift_list,
        )

        self.mapping = Image2SourceMapping(self.lens_model, self.source_model)
        self.mapping_fiducial = Image2SourceMapping(
            self.lens_model_fiducial, self.source_model
        )

        self.multi_plane_organizer = MultiPlaneOrganizer(
            lens_redshift_list=[0.2, 0.4, 0.6, 1.0],
            source_redshift_list=[0.6, 1.0, 1.0, 1.2],
            sorted_lens_redshift_index=[0, 1, 2, 3],
            sorted_source_redshift_index=[0, 1, 2, 3],
            z_lens_convention=0.2,
            z_source_convention=1.2,
            cosmo=Background(cosmo=self.cosmo),
        )

    def test_extract_a_b_factors(self):
        """Test MultiPlaneOrganizer._extract_a_b_factors()"""
        kwargs_special = {
            "factor_a_1": 1.0,
            "factor_a_2": 2.0,
            "factor_a_3": 3.0,
            "factor_a_4": 4.0,
            "factor_b_2": 5.0,
            "factor_b_3": 6.0,
        }

        a_factor_list, b_factor_list = self.multi_plane_organizer._extract_a_b_factors(
            kwargs_special
        )
        npt.assert_almost_equal(a_factor_list, [1.0, 2.0, 3.0, 4.0])
        npt.assert_almost_equal(b_factor_list, [1.0, 5.0, 6.0, 4.0])

    def test_update_lens_T_lists(self):
        """Test MultiPlaneOrganizer.update_lens_T_lists()"""
        kwargs_special = {
            "factor_a_1": 1.0,
            "factor_a_2": 1.0,
            "factor_a_3": 1.0,
            "factor_a_4": 1.0,
            "factor_b_2": 1.0,
            "factor_b_3": 1.0,
        }

        fiducial_T_z_list = copy.deepcopy(
            self.lens_model.lens_model.multi_plane_base.T_z_list
        )
        fiducial_T_ij_list = copy.deepcopy(
            self.lens_model.lens_model.multi_plane_base.T_ij_list
        )
        fiducial_T_ij_start = copy.deepcopy(self.lens_model.lens_model.T_ij_start)
        fiducial_T_ij_stop = copy.deepcopy(self.lens_model.lens_model._T_ij_stop)

        self.multi_plane_organizer.update_lens_T_lists(self.lens_model, kwargs_special)

        npt.assert_almost_equal(
            self.lens_model.lens_model.multi_plane_base.T_z_list,
            fiducial_T_z_list,
            decimal=5,
        )
        npt.assert_almost_equal(
            self.lens_model.lens_model.multi_plane_base.T_ij_list,
            fiducial_T_ij_list,
            decimal=5,
        )
        npt.assert_almost_equal(
            self.lens_model.lens_model.T_ij_start, fiducial_T_ij_start, decimal=5
        )
        npt.assert_almost_equal(
            self.lens_model.lens_model._T_ij_stop, fiducial_T_ij_stop, decimal=5
        )

    def test_update_source_mapping_T_lists(self):
        """Test MultiPlaneOrganizer.update_source_T_lists()"""
        kwargs_special = {
            "factor_a_1": 1.0,
            "factor_a_2": 1.0,
            "factor_a_3": 1.0,
            "factor_a_4": 1.0,
            "factor_b_2": 1.0,
            "factor_b_3": 1.0,
        }

        fiducial_T_ij_start_list = copy.deepcopy(self.mapping.T_ij_start_list)
        fiducial_T_ij_end_list = copy.deepcopy(self.mapping.T_ij_end_list)

        self.multi_plane_organizer.update_source_mapping_T_lists(
            self.mapping, kwargs_special
        )

        npt.assert_almost_equal(
            self.mapping.T_ij_start_list[:2], fiducial_T_ij_start_list[:2], decimal=5
        )
        assert self.mapping.T_ij_start_list[2] == None
        assert self.mapping.T_ij_start_list[3] == None

        npt.assert_almost_equal(
            self.mapping.T_ij_end_list, fiducial_T_ij_end_list, decimal=5
        )

    def test_get_element_index(self):
        """Test MultiPlaneOrganizer._get_element_index()"""
        arr = np.array([1, 2, 3, 4, 5, 6])
        index = self.multi_plane_organizer._get_element_index(arr, 3)
        assert index == 2

        with pytest.raises(ValueError):
            self.multi_plane_organizer._get_element_index(arr, 7)

    def test_get_lens_T_lists(self):
        """Test MultiPlaneOrganizer._get_lens_T_lists()"""
        kwargs_special = {
            "factor_a_1": 1.0,
            "factor_a_2": 1.0,
            "factor_a_3": 1.0,
            "factor_a_4": 1.0,
            "factor_b_2": 1.0,
            "factor_b_3": 1.0,
        }

        T_z_list, T_ij_list = self.multi_plane_organizer._get_lens_T_lists(
            kwargs_special
        )

        npt.assert_almost_equal(
            T_z_list,
            [
                751.3399926289621,
                1409.1346240906996,
                1982.1199934375895,
                2919.3681790661544,
            ],
        )

        npt.assert_almost_equal(
            T_ij_list,
            [
                751.3399926289621,
                657.7946314617373,
                572.9853693468906,
                937.2481856285648,
            ],
        )

        multi_plane_organizer = MultiPlaneOrganizer(
            lens_redshift_list=[0.2, 0.2, 0.6, 1.0],
            source_redshift_list=[0.6, 1.0, 1.0, 1.2],
            sorted_lens_redshift_index=[0, 1, 2, 3],
            sorted_source_redshift_index=[0, 1, 2, 3],
            z_lens_convention=0.2,
            z_source_convention=1.2,
            cosmo=Background(cosmo=self.cosmo),
        )

        T_z, T_ij = multi_plane_organizer._get_lens_T_lists(kwargs_special)
        assert T_ij[1] == 0

    def test_get_D_ij(self):
        """Test MultiPlaneOrganizer._get_D_ij()"""
        kwargs_special = {
            "factor_a_1": 1.0,
            "factor_a_2": 1.0,
            "factor_a_3": 1.0,
            "factor_a_4": 1.0,
            "factor_b_2": 1.0,
            "factor_b_3": 1.0,
        }

        D_ij = self.multi_plane_organizer._get_D_ij(
            self.lens_redshift_list[1], self.lens_redshift_list[2], kwargs_special
        )
        assert D_ij == 358.1158558418066

        D_ij = self.multi_plane_organizer._get_D_ij(
            self.source_redshift_list[2], self.source_redshift_list[3], kwargs_special
        )
        assert D_ij == 175.31018095788596

        # test if an error is raised if the redshifts are not in the list
        with pytest.raises(ValueError):
            D_ij = self.multi_plane_organizer._get_D_ij(0.1, 0.2, kwargs_special)

        # test if the redshifts are not consecutive
        with pytest.raises(AssertionError):
            D_ij = self.multi_plane_organizer._get_D_ij(
                self.lens_redshift_list[0], self.lens_redshift_list[3], kwargs_special
            )

    def test_get_D_i(self):
        """Test MultiPlaneOrganizer._get_D_i()"""
        kwargs_special = {
            "factor_a_1": 1.0,
            "factor_a_2": 1.0,
            "factor_a_3": 1.0,
            "factor_a_4": 1.0,
            "factor_b_2": 1.0,
            "factor_b_3": 1.0,
        }

        D_i = self.multi_plane_organizer._get_D_i(
            self.lens_redshift_list[1], kwargs_special
        )
        assert D_i == 1006.5247314933569

        D_i = self.multi_plane_organizer._get_D_i(
            self.source_redshift_list[2], kwargs_special
        )
        assert D_i == 1459.6840895330772

        # test if an error is raised if the redshifts are not in the list
        with pytest.raises(ValueError):
            D_i = self.multi_plane_organizer._get_D_i(10.0, kwargs_special)

        assert self.multi_plane_organizer._get_D_i(0, kwargs_special) == 0
        assert (
            self.multi_plane_organizer._get_D_i(
                np.max(self.source_redshift_list), kwargs_special
            )
            == self.multi_plane_organizer._D_is_list_fiducial[0]
        )

    def test_transver_distance_start_stop(self):
        """Test MultiPlaneOrganizer._transverse_distance_start_stop()"""
        kwargs_special = {
            "factor_a_1": 1.0,
            "factor_a_2": 1.0,
            "factor_a_3": 1.0,
            "factor_a_4": 1.0,
            "factor_b_2": 1.0,
            "factor_b_3": 1.0,
        }

        (
            T_ij_start,
            T_ij_stop,
        ) = self.multi_plane_organizer._transverse_distance_start_stop(
            self.lens_redshift_list[1], self.source_redshift_list[-2], kwargs_special
        )

        assert T_ij_start == 572.9853693468906
        assert T_ij_stop == 0.0

    def test_get_source_T_start_end_lists(self):
        """Test MultiPlaneOrganizer._get_source_T_start_end_lists()"""
        kwargs_special = {
            "factor_a_1": 1.0,
            "factor_a_2": 1.0,
            "factor_a_3": 1.0,
            "factor_a_4": 1.0,
            "factor_b_2": 1.0,
            "factor_b_3": 1.0,
        }

        (
            T_ij_start_list,
            T_ij_end_list,
        ) = self.multi_plane_organizer._get_source_T_start_end_lists(kwargs_special)

        npt.assert_almost_equal(
            T_ij_start_list[:2], [751.3399926289621, 937.2481856285648]
        )
        assert T_ij_start_list[2] == None
        assert T_ij_start_list[3] == None

        npt.assert_almost_equal(T_ij_end_list, [0.0, 0.0, 0.0, 385.6823981])

    def test_start_condition(self):
        """Test MultiPlaneOrganizer._start_condition()"""
        assert self.multi_plane_organizer._start_condition(True, 1.0, 1.0) == True
        assert self.multi_plane_organizer._start_condition(False, 1.0, 1.0) == False

    def test_distance_computations(self):
        """Test the distance computations between with and without distance ratio
        sampling."""
        kwargs_lens = [
            {"center_x": 0, "center_y": 0, "theta_E": 1.0},
            {"center_x": 0, "center_y": 0, "theta_E": 0.0},
            {"center_x": 0, "center_y": 0, "theta_E": 1.0},
            {},
        ]

        kwargs_source = [
            {"center_x": 0, "center_y": 0, "amp": 1, "sigma": 0.1},
            {"center_x": 0, "center_y": 0, "amp": 1, "sigma": 0.1},  #
            {"center_x": 0, "center_y": 0, "amp": 1, "sigma": 0.1},  #
            {"center_x": 0, "center_y": 0, "amp": 1, "sigma": 0.1},  #
        ]

        kwargs_special = {
            "factor_a_1": 1.0,
            "factor_a_2": 1.0,
            "factor_a_3": 1.0,
            "factor_a_4": 1.0,
            "factor_b_2": 1.0,
            "factor_b_3": 1.0,
        }

        x, y = np.meshgrid(np.arange(-4, 4, 0.05), np.arange(-4, 4, 0.05))

        image = self.mapping.image_flux_joint(
            x, y, kwargs_lens, kwargs_source, kwargs_special=kwargs_special
        )

        npt.assert_almost_equal(
            self.mapping._lens_model.lens_model.multi_plane_base.T_ij_list,
            self.mapping_fiducial._lens_model.lens_model.multi_plane_base._T_ij_list,
        )

        array_1 = np.where(
            np.array(self.mapping._T_ij_start_list) == None,
            -1,
            self.mapping._T_ij_start_list,
        )
        array_2 = np.where(
            np.array(self.mapping_fiducial._T_ij_start_list) == None,
            -1,
            self.mapping_fiducial._T_ij_start_list,
        )
        npt.assert_almost_equal(array_1, array_2)

        npt.assert_almost_equal(
            self.mapping._T_ij_end_list, self.mapping_fiducial._T_ij_end_list
        )

        npt.assert_almost_equal(
            self.mapping._lens_model.lens_model.multi_plane_base.T_z_list,
            self.mapping_fiducial._lens_model.lens_model.multi_plane_base._T_z_list,
        )

    def test_raise(self):
        with pytest.raises(ValueError):
            MultiPlaneOrganizer(
                lens_redshift_list=[0.2, 0.4, 0.6, 1.0],
                source_redshift_list=[0.6, 1.0, 1.0, 1.2],
                sorted_lens_redshift_index=[0, 1, 2, 3],
                sorted_source_redshift_index=[0, 1, 2, 3],
                z_lens_convention=0.1,
                z_source_convention=1.2,
                cosmo=Background(cosmo=self.cosmo),
            )

        with pytest.raises(ValueError):
            MultiPlaneOrganizer(
                lens_redshift_list=[0.2, 0.4, 0.6, 1.0],
                source_redshift_list=[0.6, 1.0, 1.0, 1.2],
                sorted_lens_redshift_index=[0, 1, 2, 3],
                sorted_source_redshift_index=[0, 1, 2, 3],
                z_lens_convention=0.2,
                z_source_convention=1.0,
                cosmo=Background(cosmo=self.cosmo),
            )
